@@ -52,8 +52,7 @@ static int	 format_replace(struct format_tree *, const char *, size_t,
 static void	 format_defaults_session(struct format_tree *,
 		     struct session *);
 static void	 format_defaults_client(struct format_tree *, struct client *);
-static void	 format_defaults_winlink(struct format_tree *,
-		     struct winlink *);
+static void	 format_defaults_winlink(struct format_tree *, struct winlink *);
 
 /* Entry in format job tree. */
 struct format_job {
@@ -93,7 +92,16 @@ format_job_cmp(struct format_job *fj1, struct format_job *fj2)
 #define FORMAT_TIMESTRING 0x1
 #define FORMAT_BASENAME 0x2
 #define FORMAT_DIRNAME 0x4
-#define FORMAT_SUBSTITUTE 0x8
+#define FORMAT_QUOTE 0x8
+#define FORMAT_LITERAL 0x10
+#define FORMAT_EXPAND 0x20
+#define FORMAT_EXPANDTIME 0x40
+#define FORMAT_SESSIONS 0x80
+#define FORMAT_WINDOWS 0x100
+#define FORMAT_PANES 0x200
+
+/* Limit on recursion. */
+#define FORMAT_LOOP_LIMIT 10
 
 /* Entry in format tree. */
 struct format_entry {
@@ -106,20 +114,33 @@ struct format_entry {
 
 /* Format entry tree. */
 struct format_tree {
-	struct window		*w;
-	struct winlink		*wl;
+	struct client		*c;
 	struct session		*s;
+	struct winlink		*wl;
+	struct window		*w;
 	struct window_pane	*wp;
 
+	struct cmdq_item	*item;
 	struct client		*client;
 	u_int			 tag;
 	int			 flags;
+	time_t			 time;
+	u_int			 loop;
 
 	RB_HEAD(format_entry_tree, format_entry) tree;
 };
 static int format_entry_cmp(struct format_entry *, struct format_entry *);
 RB_GENERATE_STATIC(format_entry_tree, format_entry, entry, format_entry_cmp);
 
+/* Format modifier. */
+struct format_modifier {
+	char	  modifier[3];
+	u_int	  size;
+
+	char	**argv;
+	int	  argc;
+};
+
 /* Format entry tree comparison function. */
 static int
 format_entry_cmp(struct format_entry *fe1, struct format_entry *fe2)
@@ -187,12 +208,42 @@ static const char *format_lower[] = {
 	NULL		/* z */
 };
 
+/* Is logging enabled? */
+static inline int
+format_logging(struct format_tree *ft)
+{
+	return (log_get_level() != 0 || (ft->flags & FORMAT_VERBOSE));
+}
+
+/* Log a message if verbose. */
+static void printflike(3, 4)
+format_log1(struct format_tree *ft, const char *from, const char *fmt, ...)
+{
+	va_list			 ap;
+	char			*s;
+	static const char	 spaces[] = "          ";
+
+	if (!format_logging(ft))
+		return;
+
+	va_start(ap, fmt);
+	vasprintf(&s, fmt, ap);
+	va_end(ap);
+
+	log_debug("%s: %s", from, s);
+	if (ft->item != NULL && (ft->flags & FORMAT_VERBOSE))
+		cmdq_print(ft->item, "#%.*s%s", ft->loop, spaces, s);
+
+	free(s);
+}
+#define format_log(ft, fmt, ...) format_log1(ft, __func__, fmt, ##__VA_ARGS__)
+
 /* Format job update callback. */
 static void
 format_job_update(struct job *job)
 {
-	struct format_job	*fj = job->data;
-	struct evbuffer		*evb = job->event->input;
+	struct format_job	*fj = job_get_data(job);
+	struct evbuffer		*evb = job_get_event(job)->input;
 	char			*line = NULL, *next;
 	time_t			 t;
 
@@ -221,18 +272,19 @@ format_job_update(struct job *job)
 static void
 format_job_complete(struct job *job)
 {
-	struct format_job	*fj = job->data;
+	struct format_job	*fj = job_get_data(job);
+	struct evbuffer		*evb = job_get_event(job)->input;
 	char			*line, *buf;
 	size_t			 len;
 
 	fj->job = NULL;
 
 	buf = NULL;
-	if ((line = evbuffer_readline(job->event->input)) == NULL) {
-		len = EVBUFFER_LENGTH(job->event->input);
+	if ((line = evbuffer_readline(evb)) == NULL) {
+		len = EVBUFFER_LENGTH(evb);
 		buf = xmalloc(len + 1);
 		if (len != 0)
-			memcpy(buf, EVBUFFER_DATA(job->event->input), len);
+			memcpy(buf, EVBUFFER_DATA(evb), len);
 		buf[len] = '\0';
 	} else
 		buf = line;
@@ -294,8 +346,11 @@ format_job_get(struct format_tree *ft, const char *cmd)
 		force = (ft->flags & FORMAT_FORCE);
 
 	t = time(NULL);
-	if (fj->job == NULL && (force || fj->last != t)) {
-		fj->job = job_run(expanded, NULL, NULL, format_job_update,
+	if (force && fj->job != NULL)
+	       job_free(fj->job);
+	if (force || (fj->job == NULL && fj->last != t)) {
+		fj->job = job_run(expanded, NULL,
+		    server_client_get_cwd(ft->client, NULL), format_job_update,
 		    format_job_complete, NULL, fj, JOB_NOWAIT);
 		if (fj->job == NULL) {
 			free(fj->out);
@@ -585,6 +640,8 @@ format_cb_pane_tabs(struct format_tree *ft, struct format_entry *fe)
 		return;
 
 	buffer = evbuffer_new();
+	if (buffer == NULL)
+		fatalx("out of memory");
 	for (i = 0; i < wp->base.grid->sx; i++) {
 		if (!bit_test(wp->base.tabs, i))
 			continue;
@@ -615,6 +672,8 @@ format_cb_session_group_list(struct format_tree *ft, struct format_entry *fe)
 		return;
 
 	buffer = evbuffer_new();
+	if (buffer == NULL)
+		fatalx("out of memory");
 	TAILQ_FOREACH(loop, &sg->sessions, gentry) {
 		if (EVBUFFER_LENGTH(buffer) > 0)
 			evbuffer_add(buffer, ",", 1);
@@ -625,6 +684,37 @@ format_cb_session_group_list(struct format_tree *ft, struct format_entry *fe)
 	evbuffer_free(buffer);
 }
 
+/* Callback for pane_in_mode. */
+static void
+format_cb_pane_in_mode(struct format_tree *ft, struct format_entry *fe)
+{
+	struct window_pane		*wp = ft->wp;
+	u_int				 n = 0;
+	struct window_mode_entry	*wme;
+
+	if (wp == NULL)
+		return;
+
+	TAILQ_FOREACH(wme, &wp->modes, entry)
+	    n++;
+	xasprintf(&fe->value, "%u", n);
+}
+
+/* Callback for cursor_character. */
+static void
+format_cb_cursor_character(struct format_tree *ft, struct format_entry *fe)
+{
+	struct window_pane	*wp = ft->wp;
+	struct grid_cell	 gc;
+
+	if (wp == NULL)
+		return;
+
+	grid_view_get_cell(wp->base.grid, wp->base.cx, wp->base.cy, &gc);
+	if (~gc.flags & GRID_FLAG_PADDING)
+		xasprintf(&fe->value, "%.*s", (int)gc.data.size, gc.data.data);
+}
+
 /* Merge a format tree. */
 static void
 format_merge(struct format_tree *ft, struct format_tree *from)
@@ -641,7 +731,9 @@ format_merge(struct format_tree *ft, struct format_tree *from)
 struct format_tree *
 format_create(struct client *c, struct cmdq_item *item, int tag, int flags)
 {
-	struct format_tree	*ft;
+	struct format_tree		 *ft;
+	const struct window_mode	**wm;
+	char				  tmp[64];
 
 	if (!event_initialized(&format_job_event)) {
 		evtimer_set(&format_job_event, format_job_timer, NULL);
@@ -655,9 +747,11 @@ format_create(struct client *c, struct cmdq_item *item, int tag, int flags)
 		ft->client = c;
 		ft->client->references++;
 	}
+	ft->item = item;
 
 	ft->tag = tag;
 	ft->flags = flags;
+	ft->time = time(NULL);
 
 	format_add(ft, "version", "%s", VERSION);
 	format_add_cb(ft, "host", format_cb_host);
@@ -666,6 +760,14 @@ format_create(struct client *c, struct cmdq_item *item, int tag, int flags)
 	format_add(ft, "socket_path", "%s", socket_path);
 	format_add_tv(ft, "start_time", &start_time);
 
+	for (wm = all_window_modes; *wm != NULL; wm++) {
+		if ((*wm)->default_format != NULL) {
+			xsnprintf(tmp, sizeof tmp, "%s_format", (*wm)->name);
+			tmp[strcspn(tmp, "-")] = '_';
+			format_add(ft, tmp, "%s", (*wm)->default_format);
+		}
+	}
+
 	if (item != NULL) {
 		if (item->cmd != NULL)
 			format_add(ft, "command", "%s", item->cmd->entry->name);
@@ -694,6 +796,30 @@ format_free(struct format_tree *ft)
 	free(ft);
 }
 
+/* Walk each format. */
+void
+format_each(struct format_tree *ft, void (*cb)(const char *, const char *,
+    void *), void *arg)
+{
+	struct format_entry	*fe;
+	static char		 s[64];
+
+	RB_FOREACH(fe, format_entry_tree, &ft->tree) {
+		if (fe->t != 0) {
+			xsnprintf(s, sizeof s, "%lld", (long long)fe->t);
+			cb(fe->key, fe->value, s);
+		} else {
+			if (fe->value == NULL && fe->cb != NULL) {
+				fe->cb(ft, fe);
+				if (fe->value == NULL)
+					fe->value = xstrdup("");
+			}
+			cb(fe->key, fe->value, arg);
+		}
+	}
+}
+
+
 /* Add a key-value pair. */
 void
 format_add(struct format_tree *ft, const char *key, const char *fmt, ...)
@@ -769,6 +895,23 @@ format_add_cb(struct format_tree *ft, const char *key, format_cb cb)
 	fe->value = NULL;
 }
 
+/* Quote special characters in string. */
+static char *
+format_quote(const char *s)
+{
+	const char	*cp;
+	char		*out, *at;
+
+	at = out = xmalloc(strlen(s) * 2 + 1);
+	for (cp = s; *cp != '\0'; cp++) {
+		if (strchr("|&;<>()$`\\\"'*?[# =%", *cp) != NULL)
+			*at++ = '\\';
+		*at++ = *cp;
+	}
+	*at = '\0';
+	return (out);
+}
+
 /* Find a format entry. */
 static char *
 format_find(struct format_tree *ft, const char *key, int modifiers)
@@ -851,12 +994,17 @@ found:
 		copy = xstrdup(dirname(saved));
 		free(saved);
 	}
+	if (modifiers & FORMAT_QUOTE) {
+		saved = copy;
+		copy = xstrdup(format_quote(saved));
+		free(saved);
+	}
 	return (copy);
 }
 
 /* Skip until end. */
-static const char *
-format_skip(const char *s, char end)
+const char *
+format_skip(const char *s, const char *end)
 {
 	int	brackets = 0;
 
@@ -869,7 +1017,7 @@ format_skip(const char *s, char end)
 		}
 		if (*s == '}')
 			brackets--;
-		if (*s == end && brackets == 0)
+		if (strchr(end, *s) != NULL && brackets == 0)
 			break;
 	}
 	if (*s == '\0')
@@ -879,17 +1027,27 @@ format_skip(const char *s, char end)
 
 /* Return left and right alternatives separated by commas. */
 static int
-format_choose(char *s, char **left, char **right)
+format_choose(struct format_tree *ft, const char *s, char **left, char **right,
+    int expand)
 {
-	char	*cp;
+	const char	*cp;
+	char		*left0, *right0;
 
-	cp = (char *)format_skip(s, ',');
+	cp = format_skip(s, ",");
 	if (cp == NULL)
 		return (-1);
-	*cp = '\0';
+	left0 = xstrndup(s, cp - s);
+	right0 = xstrdup(cp + 1);
 
-	*left = s;
-	*right = cp + 1;
+	if (expand) {
+		*left = format_expand(ft, left0);
+		free(left0);
+		*right = format_expand(ft, right0);
+		free(right0);
+	} else {
+		*left = left0;
+		*right = right0;
+	}
 	return (0);
 }
 
@@ -902,231 +1060,570 @@ format_true(const char *s)
 	return (0);
 }
 
+/* Check if modifier end. */
+static int
+format_is_end(char c)
+{
+	return (c == ';' || c == ':');
+}
+
+/* Add to modifier list. */
+static void
+format_add_modifier(struct format_modifier **list, u_int *count,
+    const char *c, size_t n, char **argv, int argc)
+{
+	struct format_modifier *fm;
+
+	*list = xreallocarray(*list, (*count) + 1, sizeof **list);
+	fm = &(*list)[(*count)++];
+
+	memcpy(fm->modifier, c, n);
+	fm->modifier[n] = '\0';
+	fm->size = n;
+
+	fm->argv = argv;
+	fm->argc = argc;
+}
+
+/* Free modifier list. */
+static void
+format_free_modifiers(struct format_modifier *list, u_int count)
+{
+	u_int	i;
+
+	for (i = 0; i < count; i++)
+		cmd_free_argv(list[i].argc, list[i].argv);
+	free(list);
+}
+
+/* Build modifier list. */
+static struct format_modifier *
+format_build_modifiers(struct format_tree *ft, const char **s, u_int *count)
+{
+	const char		*cp = *s, *end;
+	struct format_modifier	*list = NULL;
+	char			 c, last[] = "X;:", **argv, *value;
+	int			 argc;
+
+	/*
+	 * Modifiers are a ; separated list of the forms:
+	 *      l,m,C,b,d,t,q,E,T,S,W,P
+	 *	=a
+	 *	=/a
+	 *      =/a/
+	 *	s/a/b/
+	 *	s/a/b
+	 *	||,&&,!=,==
+	 */
+
+	*count = 0;
+
+	while (*cp != '\0' && *cp != ':') {
+		/* Skip and separator character. */
+		if (*cp == ';')
+			cp++;
+
+		/* Check single character modifiers with no arguments. */
+		if (strchr("lmCbdtqETSWP", cp[0]) != NULL &&
+		    format_is_end(cp[1])) {
+			format_add_modifier(&list, count, cp, 1, NULL, 0);
+			cp++;
+			continue;
+		}
+
+		/* Then try double character with no arguments. */
+		if ((memcmp("||", cp, 2) == 0 ||
+		    memcmp("&&", cp, 2) == 0 ||
+		    memcmp("!=", cp, 2) == 0 ||
+		    memcmp("==", cp, 2) == 0) &&
+		    format_is_end(cp[2])) {
+			format_add_modifier(&list, count, cp, 2, NULL, 0);
+			cp += 2;
+			continue;
+		}
+
+		/* Now try single character with arguments. */
+		if (strchr("s=", cp[0]) == NULL)
+			break;
+		c = cp[0];
+
+		/* No arguments provided. */
+		if (format_is_end(cp[1])) {
+			format_add_modifier(&list, count, cp, 1, NULL, 0);
+			cp++;
+			continue;
+		}
+		argv = NULL;
+		argc = 0;
+
+		/* Single argument with no wrapper character. */
+		if (!ispunct(cp[1]) || cp[1] == '-') {
+			end = format_skip(cp + 1, ":;");
+			if (end == NULL)
+				break;
+
+			argv = xcalloc(1, sizeof *argv);
+			value = xstrndup(cp + 1, end - (cp + 1));
+			argv[0] = format_expand(ft, value);
+			free(value);
+			argc = 1;
+
+			format_add_modifier(&list, count, &c, 1, argv, argc);
+			cp = end;
+			continue;
+		}
+
+		/* Multiple arguments with a wrapper character. */
+		last[0] = cp[1];
+		cp++;
+		do {
+			if (cp[0] == last[0] && format_is_end(cp[1])) {
+				cp++;
+				break;
+			}
+			end = format_skip(cp + 1, last);
+			if (end == NULL)
+				break;
+			cp++;
+
+			argv = xreallocarray (argv, argc + 1, sizeof *argv);
+			value = xstrndup(cp, end - cp);
+			argv[argc++] = format_expand(ft, value);
+			free(value);
+
+			cp = end;
+		} while (!format_is_end(cp[0]));
+		format_add_modifier(&list, count, &c, 1, argv, argc);
+	}
+	if (*cp != ':') {
+		format_free_modifiers(list, *count);
+		*count = 0;
+		return (NULL);
+	}
+	*s = cp + 1;
+	return list;
+}
+
+/* Perform substitution in string. */
+static char *
+format_substitute(const char *source, const char *from, const char *to)
+{
+	char		*copy, *new;
+	const char	*cp;
+	size_t		 fromlen, tolen, newlen, used;
+
+	fromlen = strlen(from);
+	tolen = strlen(to);
+
+	newlen = strlen(source) + 1;
+	copy = new = xmalloc(newlen);
+
+	for (cp = source; *cp != '\0'; /* nothing */) {
+		if (strncmp(cp, from, fromlen) != 0) {
+			*new++ = *cp++;
+			continue;
+		}
+		used = new - copy;
+
+		newlen += tolen;
+		copy = xrealloc(copy, newlen);
+
+		new = copy + used;
+		memcpy(new, to, tolen);
+
+		new += tolen;
+		cp += fromlen;
+	}
+
+	*new = '\0';
+	return (copy);
+}
+
+/* Loop over sessions. */
+static char *
+format_loop_sessions(struct format_tree *ft, const char *fmt)
+{
+	struct client		*c = ft->client;
+	struct cmdq_item	*item = ft->item;
+	struct format_tree	*nft;
+	char			*expanded, *value;
+	size_t			 valuelen;
+	struct session		*s;
+
+	value = xcalloc(1, 1);
+	valuelen = 1;
+
+	RB_FOREACH(s, sessions, &sessions) {
+		format_log(ft, "session loop: $%u", s->id);
+		nft = format_create(c, item, FORMAT_NONE, ft->flags);
+		nft->loop = ft->loop;
+		format_defaults(nft, ft->c, s, NULL, NULL);
+		expanded = format_expand(nft, fmt);
+		format_free(nft);
+
+		valuelen += strlen(expanded);
+		value = xrealloc(value, valuelen);
+
+		strlcat(value, expanded, valuelen);
+		free(expanded);
+	}
+
+	return (value);
+}
+
+/* Loop over windows. */
+static char *
+format_loop_windows(struct format_tree *ft, const char *fmt)
+{
+	struct client		*c = ft->client;
+	struct cmdq_item	*item = ft->item;
+	struct format_tree	*nft;
+	char			*all, *active, *use, *expanded, *value;
+	size_t			 valuelen;
+	struct winlink		*wl;
+	struct window		*w;
+
+	if (ft->s == NULL) {
+		format_log(ft, "window loop but no session");
+		return (NULL);
+	}
+
+	if (format_choose(ft, fmt, &all, &active, 0) != 0) {
+		all = xstrdup(fmt);
+		active = NULL;
+	}
+
+	value = xcalloc(1, 1);
+	valuelen = 1;
+
+	RB_FOREACH(wl, winlinks, &ft->s->windows) {
+		w = wl->window;
+		format_log(ft, "window loop: %u @%u", wl->idx, w->id);
+		if (active != NULL && wl == ft->s->curw)
+			use = active;
+		else
+			use = all;
+		nft = format_create(c, item, FORMAT_WINDOW|w->id, ft->flags);
+		nft->loop = ft->loop;
+		format_defaults(nft, ft->c, ft->s, wl, NULL);
+		expanded = format_expand(nft, use);
+		format_free(nft);
+
+		valuelen += strlen(expanded);
+		value = xrealloc(value, valuelen);
+
+		strlcat(value, expanded, valuelen);
+		free(expanded);
+	}
+
+	free(active);
+	free(all);
+
+	return (value);
+}
+
+/* Loop over panes. */
+static char *
+format_loop_panes(struct format_tree *ft, const char *fmt)
+{
+	struct client		*c = ft->client;
+	struct cmdq_item	*item = ft->item;
+	struct format_tree	*nft;
+	char			*all, *active, *use, *expanded, *value;
+	size_t			 valuelen;
+	struct window_pane	*wp;
+
+	if (ft->w == NULL) {
+		format_log(ft, "pane loop but no window");
+		return (NULL);
+	}
+
+	if (format_choose(ft, fmt, &all, &active, 0) != 0) {
+		all = xstrdup(fmt);
+		active = NULL;
+	}
+
+	value = xcalloc(1, 1);
+	valuelen = 1;
+
+	TAILQ_FOREACH(wp, &ft->w->panes, entry) {
+		format_log(ft, "pane loop: %%%u", wp->id);
+		if (active != NULL && wp == ft->w->active)
+			use = active;
+		else
+			use = all;
+		nft = format_create(c, item, FORMAT_PANE|wp->id, ft->flags);
+		nft->loop = ft->loop;
+		format_defaults(nft, ft->c, ft->s, ft->wl, wp);
+		expanded = format_expand(nft, use);
+		format_free(nft);
+
+		valuelen += strlen(expanded);
+		value = xrealloc(value, valuelen);
+
+		strlcat(value, expanded, valuelen);
+		free(expanded);
+	}
+
+	free(active);
+	free(all);
+
+	return (value);
+}
+
 /* Replace a key. */
 static int
 format_replace(struct format_tree *ft, const char *key, size_t keylen,
     char **buf, size_t *len, size_t *off)
 {
 	struct window_pane	*wp = ft->wp;
-	char			*copy, *copy0, *endptr, *ptr, *found, *new, sep;
-	char			*value, *from = NULL, *to = NULL, *left, *right;
-	size_t			 valuelen, newlen, fromlen, tolen, used;
-	long			 limit = 0;
-	int			 modifiers = 0, compare = 0, search = 0;
-	int			 literal = 0;
+	const char		*errptr, *copy, *cp;
+	char			*copy0, *condition, *found, *new;
+	char			*value, *left, *right;
+	size_t			 valuelen;
+	int			 modifiers = 0, limit = 0;
+	struct format_modifier  *list, *fm, *cmp = NULL, *search = NULL;
+	struct format_modifier  *sub = NULL;
+	u_int			 i, count;
+	int			 j;
 
 	/* Make a copy of the key. */
-	copy0 = copy = xmalloc(keylen + 1);
-	memcpy(copy, key, keylen);
-	copy[keylen] = '\0';
+	copy = copy0 = xstrndup(key, keylen);
 
-	/* Is there a length limit or whatnot? */
-	switch (copy[0]) {
-	case 'l':
-		if (copy[1] != ':')
-			break;
-		literal = 1;
-		copy += 2;
-		break;
-	case 'm':
-		if (copy[1] != ':')
-			break;
-		compare = -2;
-		copy += 2;
-		break;
-	case 'C':
-		if (copy[1] != ':')
-			break;
-		search = 1;
-		copy += 2;
-		break;
-	case '|':
-		if (copy[1] != '|' || copy[2] != ':')
-			break;
-		compare = -3;
-		copy += 3;
-		break;
-	case '&':
-		if (copy[1] != '&' || copy[2] != ':')
-			break;
-		compare = -4;
-		copy += 3;
-		break;
-	case '!':
-		if (copy[1] == '=' && copy[2] == ':') {
-			compare = -1;
-			copy += 3;
-			break;
+	/* Process modifier list. */
+	list = format_build_modifiers(ft, &copy, &count);
+	for (i = 0; i < count; i++) {
+		fm = &list[i];
+		if (format_logging(ft)) {
+			format_log(ft, "modifier %u is %s", i, fm->modifier);
+			for (j = 0; j < fm->argc; j++) {
+				format_log(ft, "modifier %u argument %d: %s", i,
+				    j, fm->argv[j]);
+			}
 		}
-		break;
-	case '=':
-		if (copy[1] == '=' && copy[2] == ':') {
-			compare = 1;
-			copy += 3;
-			break;
+		if (fm->size == 1) {
+			switch (fm->modifier[0]) {
+			case 'm':
+				cmp = fm;
+				break;
+			case 'C':
+				search = fm;
+				break;
+			case 's':
+				if (fm->argc != 2)
+					break;
+				sub = fm;
+				break;
+			case '=':
+				if (fm->argc != 1)
+					break;
+				limit = strtonum(fm->argv[0], INT_MIN, INT_MAX,
+				    &errptr);
+				if (errptr != NULL)
+					limit = 0;
+				break;
+			case 'l':
+				modifiers |= FORMAT_LITERAL;
+				break;
+			case 'b':
+				modifiers |= FORMAT_BASENAME;
+				break;
+			case 'd':
+				modifiers |= FORMAT_DIRNAME;
+				break;
+			case 't':
+				modifiers |= FORMAT_TIMESTRING;
+				break;
+			case 'q':
+				modifiers |= FORMAT_QUOTE;
+				break;
+			case 'E':
+				modifiers |= FORMAT_EXPAND;
+				break;
+			case 'T':
+				modifiers |= FORMAT_EXPANDTIME;
+				break;
+			case 'S':
+				modifiers |= FORMAT_SESSIONS;
+				break;
+			case 'W':
+				modifiers |= FORMAT_WINDOWS;
+				break;
+			case 'P':
+				modifiers |= FORMAT_PANES;
+				break;
+			}
+		} else if (fm->size == 2) {
+			if (strcmp(fm->modifier, "||") == 0 ||
+			    strcmp(fm->modifier, "&&") == 0 ||
+			    strcmp(fm->modifier, "==") == 0 ||
+			    strcmp(fm->modifier, "!=") == 0)
+				cmp = fm;
 		}
-		errno = 0;
-		limit = strtol(copy + 1, &endptr, 10);
-		if (errno == ERANGE && (limit == LONG_MIN || limit == LONG_MAX))
-			break;
-		if (*endptr != ':')
-			break;
-		copy = endptr + 1;
-		break;
-	case 'b':
-		if (copy[1] != ':')
-			break;
-		modifiers |= FORMAT_BASENAME;
-		copy += 2;
-		break;
-	case 'd':
-		if (copy[1] != ':')
-			break;
-		modifiers |= FORMAT_DIRNAME;
-		copy += 2;
-		break;
-	case 't':
-		if (copy[1] != ':')
-			break;
-		modifiers |= FORMAT_TIMESTRING;
-		copy += 2;
-		break;
-	case 's':
-		sep = copy[1];
-		if (sep == ':' || !ispunct((u_char)sep))
-			break;
-		from = copy + 2;
-		for (copy = from; *copy != '\0' && *copy != sep; copy++)
-			/* nothing */;
-		if (copy[0] != sep || copy == from) {
-			copy = copy0;
-			break;
-		}
-		copy[0] = '\0';
-		to = copy + 1;
-		for (copy = to; *copy != '\0' && *copy != sep; copy++)
-			/* nothing */;
-		if (copy[0] != sep || copy[1] != ':') {
-			copy = copy0;
-			break;
-		}
-		copy[0] = '\0';
-
-		modifiers |= FORMAT_SUBSTITUTE;
-		copy += 2;
-		break;
 	}
 
 	/* Is this a literal string? */
-	if (literal) {
+	if (modifiers & FORMAT_LITERAL) {
 		value = xstrdup(copy);
 		goto done;
 	}
 
-	/* Is this a comparison or a conditional? */
-	if (search) {
-		/* Search in pane. */
-		if (wp == NULL)
-			value = xstrdup("0");
-		else
-			xasprintf(&value, "%u", window_pane_search(wp, copy));
-	} else if (compare != 0) {
-		/* Comparison: compare comma-separated left and right. */
-		if (format_choose(copy, &left, &right) != 0)
+	/* Is this a loop, comparison or condition? */
+	if (modifiers & FORMAT_SESSIONS) {
+		value = format_loop_sessions(ft, copy);
+		if (value == NULL)
 			goto fail;
-		left = format_expand(ft, left);
-		right = format_expand(ft, right);
-		if (compare == -3 &&
-		    (format_true(left) || format_true(right)))
-			value = xstrdup("1");
-		else if (compare == -4 &&
-		    (format_true(left) && format_true(right)))
-			value = xstrdup("1");
-		else if (compare == 1 && strcmp(left, right) == 0)
-			value = xstrdup("1");
-		else if (compare == -1 && strcmp(left, right) != 0)
-			value = xstrdup("1");
-		else if (compare == -2 && fnmatch(left, right, 0) == 0)
-			value = xstrdup("1");
-		else
+	} else if (modifiers & FORMAT_WINDOWS) {
+		value = format_loop_windows(ft, copy);
+		if (value == NULL)
+			goto fail;
+	} else if (modifiers & FORMAT_PANES) {
+		value = format_loop_panes(ft, copy);
+		if (value == NULL)
+			goto fail;
+	} else if (search != NULL) {
+		/* Search in pane. */
+		if (wp == NULL) {
+			format_log(ft, "search '%s' but no pane", copy);
 			value = xstrdup("0");
+		} else {
+			format_log(ft, "search '%s' pane %%%u", copy,  wp->id);
+			xasprintf(&value, "%u", window_pane_search(wp, copy));
+		}
+	} else if (cmp != NULL) {
+		/* Comparison of left and right. */
+		if (format_choose(ft, copy, &left, &right, 1) != 0) {
+			format_log(ft, "compare %s syntax error: %s",
+			    cmp->modifier, copy);
+			goto fail;
+		}
+		format_log(ft, "compare %s left is: %s", cmp->modifier, left);
+		format_log(ft, "compare %s right is: %s", cmp->modifier, right);
+
+		if (strcmp(cmp->modifier, "||") == 0) {
+			if (format_true(left) || format_true(right))
+				value = xstrdup("1");
+			else
+				value = xstrdup("0");
+		} else if (strcmp(cmp->modifier, "&&") == 0) {
+			if (format_true(left) && format_true(right))
+				value = xstrdup("1");
+			else
+				value = xstrdup("0");
+		} else if (strcmp(cmp->modifier, "==") == 0) {
+			if (strcmp(left, right) == 0)
+				value = xstrdup("1");
+			else
+				value = xstrdup("0");
+		} else if (strcmp(cmp->modifier, "!=") == 0) {
+			if (strcmp(left, right) != 0)
+				value = xstrdup("1");
+			else
+				value = xstrdup("0");
+		}
+		else if (strcmp(cmp->modifier, "m") == 0) {
+			if (fnmatch(left, right, 0) == 0)
+				value = xstrdup("1");
+			else
+				value = xstrdup("0");
+		}
+
 		free(right);
 		free(left);
 	} else if (*copy == '?') {
 		/* Conditional: check first and choose second or third. */
-		ptr = (char *)format_skip(copy, ',');
-		if (ptr == NULL)
+		cp = format_skip(copy + 1, ",");
+		if (cp == NULL) {
+			format_log(ft, "condition syntax error: %s", copy + 1);
 			goto fail;
-		*ptr = '\0';
+		}
+		condition = xstrndup(copy + 1, cp - (copy + 1));
+		format_log(ft, "condition is: %s", condition);
 
-		found = format_find(ft, copy + 1, modifiers);
+		found = format_find(ft, condition, modifiers);
 		if (found == NULL) {
 			/*
-			 * If the conditional not found, try to expand it. If
+			 * If the condition not found, try to expand it. If
 			 * the expansion doesn't have any effect, then assume
 			 * false.
 			 */
-			found = format_expand(ft, copy + 1);
-			if (strcmp(found, copy + 1) == 0) {
+			found = format_expand(ft, condition);
+			if (strcmp(found, condition) == 0) {
 				free(found);
 				found = xstrdup("");
+				format_log(ft, "condition '%s' found: %s",
+				    condition, found);
+			} else {
+				format_log(ft,
+				    "condition '%s' not found; assuming false",
+				    condition);
 			}
-		}
-		if (format_choose(ptr + 1, &left, &right) != 0)
-			goto fail;
+		} else
+			format_log(ft, "condition '%s' found", condition);
 
-		if (format_true(found))
+		if (format_choose(ft, cp + 1, &left, &right, 0) != 0) {
+			format_log(ft, "condition '%s' syntax error: %s",
+			    condition, cp + 1);
+			free(found);
+			goto fail;
+		}
+		if (format_true(found)) {
+			format_log(ft, "condition '%s' is true", condition);
 			value = format_expand(ft, left);
-		else
+		} else {
+			format_log(ft, "condition '%s' is false", condition);
 			value = format_expand(ft, right);
+		}
+		free(right);
+		free(left);
+
+		free(condition);
 		free(found);
 	} else {
 		/* Neither: look up directly. */
 		value = format_find(ft, copy, modifiers);
-		if (value == NULL)
+		if (value == NULL) {
+			format_log(ft, "format '%s' not found", copy);
 			value = xstrdup("");
+		} else
+			format_log(ft, "format '%s' found: %s", copy, value);
+	}
+
+done:
+	/* Expand again if required. */
+	if (modifiers & FORMAT_EXPAND) {
+		new = format_expand(ft, value);
+		free(value);
+		value = new;
+	}
+	else if (modifiers & FORMAT_EXPANDTIME) {
+		new = format_expand_time(ft, value);
+		free(value);
+		value = new;
 	}
 
 	/* Perform substitution if any. */
-	if (modifiers & FORMAT_SUBSTITUTE) {
-		fromlen = strlen(from);
-		tolen = strlen(to);
-
-		newlen = strlen(value) + 1;
-		copy = new = xmalloc(newlen);
-		for (ptr = value; *ptr != '\0'; /* nothing */) {
-			if (strncmp(ptr, from, fromlen) != 0) {
-				*new++ = *ptr++;
-				continue;
-			}
-			used = new - copy;
-
-			newlen += tolen;
-			copy = xrealloc(copy, newlen);
-
-			new = copy + used;
-			memcpy(new, to, tolen);
-
-			new += tolen;
-			ptr += fromlen;
-		}
-		*new = '\0';
+	if (sub != NULL) {
+		new = format_substitute(value, sub->argv[0], sub->argv[1]);
+		format_log(ft, "substituted '%s' to '%s: %s", sub->argv[0],
+		    sub->argv[1], new);
 		free(value);
-		value = copy;
+		value = new;
 	}
 
 	/* Truncate the value if needed. */
 	if (limit > 0) {
-		new = utf8_trimcstr(value, limit);
+		new = format_trim_left(value, limit);
+		format_log(ft, "applied length limit %d: %s", limit, new);
 		free(value);
 		value = new;
 	} else if (limit < 0) {
-		new = utf8_rtrimcstr(value, -limit);
+		new = format_trim_right(value, -limit);
+		format_log(ft, "applied length limit %d: %s", limit, new);
 		free(value);
 		value = new;
 	}
 
-done:
 	/* Expand the buffer and copy in the value. */
 	valuelen = strlen(value);
 	while (*len - *off < valuelen + 1) {
@@ -1136,44 +1633,50 @@ done:
 	memcpy(*buf + *off, value, valuelen);
 	*off += valuelen;
 
+	format_log(ft, "replaced '%s' with '%s'", copy0, value);
 	free(value);
+
+	format_free_modifiers(list, count);
 	free(copy0);
 	return (0);
 
 fail:
+	format_log(ft, "failed %s", copy0);
+	format_free_modifiers(list, count);
 	free(copy0);
 	return (-1);
 }
 
-/* Expand keys in a template, passing through strftime first. */
-char *
-format_expand_time(struct format_tree *ft, const char *fmt, time_t t)
+/* Expand keys in a template. */
+static char *
+format_expand1(struct format_tree *ft, const char *fmt, int time)
 {
+	char		*buf, *out, *name;
+	const char	*ptr, *s;
+	size_t		 off, len, n, outlen;
+	int     	 ch, brackets;
 	struct tm	*tm;
-	char		 s[2048];
+	char		 expanded[8192];
 
 	if (fmt == NULL || *fmt == '\0')
 		return (xstrdup(""));
 
-	tm = localtime(&t);
-
-	if (strftime(s, sizeof s, fmt, tm) == 0)
+	if (ft->loop == FORMAT_LOOP_LIMIT)
 		return (xstrdup(""));
+	ft->loop++;
 
-	return (format_expand(ft, s));
-}
+	format_log(ft, "expanding format: %s", fmt);
 
-/* Expand keys in a template. */
-char *
-format_expand(struct format_tree *ft, const char *fmt)
-{
-	char		*buf, *out, *name;
-	const char	*ptr, *s, *saved = fmt;
-	size_t		 off, len, n, outlen;
-	int     	 ch, brackets;
-
-	if (fmt == NULL)
-		return (xstrdup(""));
+	if (time) {
+		tm = localtime(&ft->time);
+		if (strftime(expanded, sizeof expanded, fmt, tm) == 0) {
+			format_log(ft, "format is too long");
+			return (xstrdup(""));
+		}
+		if (format_logging(ft) && strcmp(expanded, fmt) != 0)
+			format_log(ft, "after time expanded: %s", expanded);
+		fmt = expanded;
+	}
 
 	len = 64;
 	buf = xmalloc(len);
@@ -1190,7 +1693,7 @@ format_expand(struct format_tree *ft, const char *fmt)
 		}
 		fmt++;
 
-		ch = (u_char) *fmt++;
+		ch = (u_char)*fmt++;
 		switch (ch) {
 		case '(':
 			brackets = 1;
@@ -1204,15 +1707,19 @@ format_expand(struct format_tree *ft, const char *fmt)
 				break;
 			n = ptr - fmt;
 
-			if (ft->flags & FORMAT_NOJOBS)
-				out = xstrdup("");
-			else {
-				name = xstrndup(fmt, n);
-				out = format_job_get(ft, name);
-				free(name);
-			}
-			outlen = strlen(out);
+			name = xstrndup(fmt, n);
+			format_log(ft, "found #(): %s", name);
 
+			if (ft->flags & FORMAT_NOJOBS) {
+				out = xstrdup("");
+				format_log(ft, "#() is disabled");
+			} else {
+				out = format_job_get(ft, name);
+				format_log(ft, "#() result: %s", out);
+			}
+			free(name);
+
+			outlen = strlen(out);
 			while (len - off < outlen + 1) {
 				buf = xreallocarray(buf, 2, len);
 				len *= 2;
@@ -1225,11 +1732,12 @@ format_expand(struct format_tree *ft, const char *fmt)
 			fmt += n + 1;
 			continue;
 		case '{':
-			ptr = format_skip(fmt - 2, '}');
+			ptr = format_skip((char *)fmt - 2, "}");
 			if (ptr == NULL)
 				break;
 			n = ptr - fmt;
 
+			format_log(ft, "found #{}: %.*s", (int)n, fmt);
 			if (format_replace(ft, fmt, n, &buf, &len, &off) != 0)
 				break;
 			fmt += n + 1;
@@ -1237,6 +1745,7 @@ format_expand(struct format_tree *ft, const char *fmt)
 		case '}':
 		case '#':
 		case ',':
+			format_log(ft, "found #%c", ch);
 			while (len - off < 2) {
 				buf = xreallocarray(buf, 2, len);
 				len *= 2;
@@ -1259,6 +1768,7 @@ format_expand(struct format_tree *ft, const char *fmt)
 				continue;
 			}
 			n = strlen(s);
+			format_log(ft, "found #%c: %s", ch, s);
 			if (format_replace(ft, s, n, &buf, &len, &off) != 0)
 				break;
 			continue;
@@ -1268,10 +1778,26 @@ format_expand(struct format_tree *ft, const char *fmt)
 	}
 	buf[off] = '\0';
 
-	log_debug("format '%s' -> '%s'", saved, buf);
+	format_log(ft, "result is: %s", buf);
+	ft->loop--;
+
 	return (buf);
 }
 
+/* Expand keys in a template, passing through strftime first. */
+char *
+format_expand_time(struct format_tree *ft, const char *fmt)
+{
+	return (format_expand1(ft, fmt, 1));
+}
+
+/* Expand keys in a template. */
+char *
+format_expand(struct format_tree *ft, const char *fmt)
+{
+	return (format_expand1(ft, fmt, 0));
+}
+
 /* Expand a single string. */
 char *
 format_single(struct cmdq_item *item, const char *fmt, struct client *c,
@@ -1330,8 +1856,6 @@ format_defaults_session(struct format_tree *ft, struct session *s)
 
 	format_add(ft, "session_name", "%s", s->name);
 	format_add(ft, "session_windows", "%u", winlink_count(&s->windows));
-	format_add(ft, "session_width", "%u", s->sx);
-	format_add(ft, "session_height", "%u", s->sy);
 	format_add(ft, "session_id", "$%u", s->id);
 
 	sg = session_group_contains(s);
@@ -1366,6 +1890,7 @@ format_defaults_client(struct format_tree *ft, struct client *c)
 
 	if (ft->s == NULL)
 		ft->s = c->session;
+	ft->c = c;
 
 	format_add(ft, "client_name", "%s", c->name);
 	format_add(ft, "client_pid", "%ld", (long) c->pid);
@@ -1434,8 +1959,11 @@ format_defaults_window(struct format_tree *ft, struct window *w)
 static void
 format_defaults_winlink(struct format_tree *ft, struct winlink *wl)
 {
+	struct client	*c = ft->c;
 	struct session	*s = wl->session;
 	struct window	*w = wl->window;
+	int		 flag;
+	u_int		 ox, oy, sx, sy;
 
 	if (ft->w == NULL)
 		ft->w = wl->window;
@@ -1443,11 +1971,25 @@ format_defaults_winlink(struct format_tree *ft, struct winlink *wl)
 
 	format_defaults_window(ft, w);
 
+	if (c != NULL) {
+		flag = tty_window_offset(&c->tty, &ox, &oy, &sx, &sy);
+		format_add(ft, "window_bigger", "%d", flag);
+		if (flag) {
+			format_add(ft, "window_offset_x", "%u", ox);
+			format_add(ft, "window_offset_y", "%u", oy);
+		}
+	}
+
 	format_add(ft, "window_index", "%d", wl->idx);
 	format_add_cb(ft, "window_stack_index", format_cb_window_stack_index);
 	format_add(ft, "window_flags", "%s", window_printable_flags(wl));
 	format_add(ft, "window_active", "%d", wl == s->curw);
 
+	format_add(ft, "window_start_flag", "%d",
+	    !!(wl == RB_MIN(winlinks, &s->windows)));
+	format_add(ft, "window_end_flag", "%d",
+	    !!(wl == RB_MAX(winlinks, &s->windows)));
+
 	format_add(ft, "window_bell_flag", "%d",
 	    !!(wl->flags & WINLINK_BELL));
 	format_add(ft, "window_activity_flag", "%d",
@@ -1463,12 +2005,14 @@ format_defaults_winlink(struct format_tree *ft, struct winlink *wl)
 void
 format_defaults_pane(struct format_tree *ft, struct window_pane *wp)
 {
-	struct grid	*gd = wp->base.grid;
-	int  		 status = wp->status;
-	u_int		 idx;
+	struct window			*w = wp->window;
+	struct grid			*gd = wp->base.grid;
+	int  				 status = wp->status;
+	u_int				 idx;
+	struct window_mode_entry	*wme;
 
 	if (ft->w == NULL)
-		ft->w = wp->window;
+		ft->w = w;
 	ft->wp = wp;
 
 	format_add(ft, "history_size", "%u", gd->hsize);
@@ -1483,7 +2027,7 @@ format_defaults_pane(struct format_tree *ft, struct window_pane *wp)
 	format_add(ft, "pane_height", "%u", wp->sy);
 	format_add(ft, "pane_title", "%s", wp->base.title);
 	format_add(ft, "pane_id", "%%%u", wp->id);
-	format_add(ft, "pane_active", "%d", wp == wp->window->active);
+	format_add(ft, "pane_active", "%d", wp == w->active);
 	format_add(ft, "pane_input_off", "%d", !!(wp->flags & PANE_INPUTOFF));
 	format_add(ft, "pane_pipe", "%d", wp->pipe_fd != -1);
 
@@ -1491,25 +2035,25 @@ format_defaults_pane(struct format_tree *ft, struct window_pane *wp)
 		format_add(ft, "pane_dead_status", "%d", WEXITSTATUS(status));
 	format_add(ft, "pane_dead", "%d", wp->fd == -1);
 
-	if (window_pane_visible(wp)) {
-		format_add(ft, "pane_left", "%u", wp->xoff);
-		format_add(ft, "pane_top", "%u", wp->yoff);
-		format_add(ft, "pane_right", "%u", wp->xoff + wp->sx - 1);
-		format_add(ft, "pane_bottom", "%u", wp->yoff + wp->sy - 1);
-		format_add(ft, "pane_at_left", "%d", wp->xoff == 0);
-		format_add(ft, "pane_at_top", "%d", wp->yoff == 0);
-		format_add(ft, "pane_at_right", "%d",
-		    wp->xoff + wp->sx == wp->window->sx);
-		format_add(ft, "pane_at_bottom", "%d",
-		    wp->yoff + wp->sy == wp->window->sy);
-	}
+	format_add(ft, "pane_left", "%u", wp->xoff);
+	format_add(ft, "pane_top", "%u", wp->yoff);
+	format_add(ft, "pane_right", "%u", wp->xoff + wp->sx - 1);
+	format_add(ft, "pane_bottom", "%u", wp->yoff + wp->sy - 1);
+	format_add(ft, "pane_at_left", "%d", wp->xoff == 0);
+	format_add(ft, "pane_at_top", "%d", wp->yoff == 0);
+	format_add(ft, "pane_at_right", "%d", wp->xoff + wp->sx == w->sx);
+	format_add(ft, "pane_at_bottom", "%d", wp->yoff + wp->sy == w->sy);
 
-	format_add(ft, "pane_in_mode", "%d", wp->screen != &wp->base);
-	if (wp->mode != NULL)
-		format_add(ft, "pane_mode", "%s", wp->mode->name);
+	wme = TAILQ_FIRST(&wp->modes);
+	if (wme != NULL) {
+		format_add(ft, "pane_mode", "%s", wme->mode->name);
+		if (wme->mode->formats != NULL)
+			wme->mode->formats(wme, ft);
+	}
+	format_add_cb(ft, "pane_in_mode", format_cb_pane_in_mode);
 
 	format_add(ft, "pane_synchronized", "%d",
-	    !!options_get_number(wp->window->options, "synchronize-panes"));
+	    !!options_get_number(w->options, "synchronize-panes"));
 	if (wp->searchstr != NULL)
 		format_add(ft, "pane_search_string", "%s", wp->searchstr);
 
@@ -1521,11 +2065,11 @@ format_defaults_pane(struct format_tree *ft, struct window_pane *wp)
 
 	format_add(ft, "cursor_x", "%u", wp->base.cx);
 	format_add(ft, "cursor_y", "%u", wp->base.cy);
+	format_add_cb(ft, "cursor_character", format_cb_cursor_character);
+
 	format_add(ft, "scroll_region_upper", "%u", wp->base.rupper);
 	format_add(ft, "scroll_region_lower", "%u", wp->base.rlower);
 
-	window_copy_add_formats(wp, ft);
-
 	format_add(ft, "alternate_on", "%d", wp->saved_grid ? 1 : 0);
 	format_add(ft, "alternate_saved_x", "%u", wp->saved_cx);
 	format_add(ft, "alternate_saved_y", "%u", wp->saved_cy);
