@@ -167,7 +167,7 @@ layout_parse(struct window *w, const char *layout)
 
 	/* Update pane offsets and sizes. */
 	layout_fix_offsets(lc);
-	layout_fix_panes(w, lc->sx, lc->sy);
+	layout_fix_panes(w);
 
 	/* Then resize the layout back to the original window size. */
 	layout_resize(w, sx, sy);
@@ -175,7 +175,7 @@ layout_parse(struct window *w, const char *layout)
 
 	layout_print_cell(lc, __func__, 0);
 
-	notify_window_layout_changed(w);
+	notify_window("window-layout-changed", w);
 
 	return (0);
 
