@@ -38,6 +38,9 @@ static const char *options_table_mode_keys_list[] = {
 static const char *options_table_clock_mode_style_list[] = {
 	"12", "24", NULL
 };
+static const char *options_table_status_list[] = {
+	"off", "on", "2", "3", "4", "5", NULL
+};
 static const char *options_table_status_keys_list[] = {
 	"emacs", "vi", NULL
 };
@@ -59,6 +62,72 @@ static const char *options_table_pane_status_list[] = {
 static const char *options_table_set_clipboard_list[] = {
 	"off", "external", "on", NULL
 };
+static const char *options_table_window_size_list[] = {
+	"largest", "smallest", "manual", NULL
+};
+
+/* Status line format. */
+#define OPTIONS_TABLE_STATUS_FORMAT1 \
+	"#[align=left range=left #{status-left-style}]" \
+	"#{T;=/#{status-left-length}:status-left}#[norange default]" \
+	"#[list=on align=#{status-justify}]" \
+	"#[list=left-marker]<#[list=right-marker]>#[list=on]" \
+	"#{W:" \
+		"#[range=window|#{window_index} " \
+			"#{window-status-style}" \
+			"#{?#{&&:#{window_last_flag}," \
+				"#{!=:#{window-status-last-style},default}}, " \
+				"#{window-status-last-style}," \
+			"}" \
+			"#{?#{&&:#{window_bell_flag}," \
+				"#{!=:#{window-status-bell-style},default}}, " \
+				"#{window-status-bell-style}," \
+				"#{?#{&&:#{||:#{window_activity_flag}," \
+					     "#{window_silence_flag}}," \
+					"#{!=:" \
+					"#{window-status-activity-style}," \
+					"default}}, " \
+					"#{window-status-activity-style}," \
+				"}" \
+			"}" \
+		"]" \
+		"#{T:window-status-format}" \
+		"#[norange default]" \
+		"#{?window_end_flag,,#{window-status-separator}}" \
+	"," \
+		"#[range=window|#{window_index} list=focus " \
+			"#{?#{!=:#{window-status-current-style},default}," \
+	                        "#{window-status-current-style}," \
+	                        "#{window-status-style}" \
+	                "}" \
+			"#{?#{&&:#{window_last_flag}," \
+				"#{!=:#{window-status-last-style},default}}, " \
+				"#{window-status-last-style}," \
+			"}" \
+			"#{?#{&&:#{window_bell_flag}," \
+				"#{!=:#{window-status-bell-style},default}}, " \
+				"#{window-status-bell-style}," \
+				"#{?#{&&:#{||:#{window_activity_flag}," \
+					     "#{window_silence_flag}}," \
+					"#{!=:" \
+					"#{window-status-activity-style}," \
+					"default}}, " \
+					"#{window-status-activity-style}," \
+				"}" \
+			"}" \
+		"]" \
+		"#{T:window-status-current-format}" \
+		"#[norange list=on default]" \
+		"#{?window_end_flag,,#{window-status-separator}}" \
+	"}" \
+	"#[nolist align=right range=right #{status-right-style}]" \
+	"#{T;=/#{status-right-length}:status-right}#[norange default]"
+#define OPTIONS_TABLE_STATUS_FORMAT2 \
+	"#[align=centre]#{P:#{?pane_active,#[reverse],}" \
+	"#{pane_index}[#{pane_width}x#{pane_height}]#[default] }"
+static const char *options_table_status_format_default[] = {
+	OPTIONS_TABLE_STATUS_FORMAT1, OPTIONS_TABLE_STATUS_FORMAT2, NULL
+};
 
 /* Top-level options. */
 const struct options_table_entry options_table[] = {
@@ -193,6 +262,13 @@ const struct options_table_entry options_table[] = {
 	  .default_str = _PATH_BSHELL
 	},
 
+	{ .name = "default-size",
+	  .type = OPTIONS_TABLE_STRING,
+	  .scope = OPTIONS_TABLE_SESSION,
+	  .pattern = "[0-9]*x[0-9]*",
+	  .default_str = "80x24"
+	},
+
 	{ .name = "destroy-unattached",
 	  .type = OPTIONS_TABLE_FLAG,
 	  .scope = OPTIONS_TABLE_SESSION,
@@ -261,54 +337,12 @@ const struct options_table_entry options_table[] = {
 	  .default_str = "lock -np"
 	},
 
-	{ .name = "message-attr",
-	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .scope = OPTIONS_TABLE_SESSION,
-	  .default_num = 0,
-	  .style = "message-style"
-	},
-
-	{ .name = "message-bg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_SESSION,
-	  .default_num = 3,
-	  .style = "message-style"
-	},
-
-	{ .name = "message-command-attr",
-	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .scope = OPTIONS_TABLE_SESSION,
-	  .default_num = 0,
-	  .style = "message-command-style"
-	},
-
-	{ .name = "message-command-bg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_SESSION,
-	  .default_num = 0,
-	  .style = "message-command-style"
-	},
-
-	{ .name = "message-command-fg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_SESSION,
-	  .default_num = 3,
-	  .style = "message-command-style"
-	},
-
 	{ .name = "message-command-style",
 	  .type = OPTIONS_TABLE_STYLE,
 	  .scope = OPTIONS_TABLE_SESSION,
 	  .default_str = "bg=black,fg=yellow"
 	},
 
-	{ .name = "message-fg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_SESSION,
-	  .default_num = 0,
-	  .style = "message-style"
-	},
-
 	{ .name = "message-style",
 	  .type = OPTIONS_TABLE_STYLE,
 	  .scope = OPTIONS_TABLE_SESSION,
@@ -367,30 +401,28 @@ const struct options_table_entry options_table[] = {
 	},
 
 	{ .name = "status",
-	  .type = OPTIONS_TABLE_FLAG,
+	  .type = OPTIONS_TABLE_CHOICE,
 	  .scope = OPTIONS_TABLE_SESSION,
+	  .choices = options_table_status_list,
 	  .default_num = 1
 	},
 
-	{ .name = "status-attr",
-	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .scope = OPTIONS_TABLE_SESSION,
-	  .default_num = 0,
-	  .style = "status-style"
-	},
-
 	{ .name = "status-bg",
 	  .type = OPTIONS_TABLE_COLOUR,
 	  .scope = OPTIONS_TABLE_SESSION,
 	  .default_num = 2,
-	  .style = "status-style"
 	},
 
 	{ .name = "status-fg",
 	  .type = OPTIONS_TABLE_COLOUR,
 	  .scope = OPTIONS_TABLE_SESSION,
 	  .default_num = 0,
-	  .style = "status-style"
+	},
+
+	{ .name = "status-format",
+	  .type = OPTIONS_TABLE_ARRAY,
+	  .scope = OPTIONS_TABLE_SESSION,
+	  .default_arr = options_table_status_format_default,
 	},
 
 	{ .name = "status-interval",
@@ -421,27 +453,6 @@ const struct options_table_entry options_table[] = {
 	  .default_str = "[#S] "
 	},
 
-	{ .name = "status-left-attr",
-	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .scope = OPTIONS_TABLE_SESSION,
-	  .default_num = 0,
-	  .style = "status-left-style"
-	},
-
-	{ .name = "status-left-bg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_SESSION,
-	  .default_num = 8,
-	  .style = "status-left-style"
-	},
-
-	{ .name = "status-left-fg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_SESSION,
-	  .default_num = 8,
-	  .style = "status-left-style"
-	},
-
 	{ .name = "status-left-length",
 	  .type = OPTIONS_TABLE_NUMBER,
 	  .scope = OPTIONS_TABLE_SESSION,
@@ -466,28 +477,9 @@ const struct options_table_entry options_table[] = {
 	{ .name = "status-right",
 	  .type = OPTIONS_TABLE_STRING,
 	  .scope = OPTIONS_TABLE_SESSION,
-	  .default_str = " \"#{=21:pane_title}\" %H:%M %d-%b-%y"
-	},
-
-	{ .name = "status-right-attr",
-	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .scope = OPTIONS_TABLE_SESSION,
-	  .default_num = 0,
-	  .style = "status-right-style"
-	},
-
-	{ .name = "status-right-bg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_SESSION,
-	  .default_num = 8,
-	  .style = "status-right-style"
-	},
-
-	{ .name = "status-right-fg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_SESSION,
-	  .default_num = 8,
-	  .style = "status-right-style"
+	  .default_str = "#{?window_bigger,"
+	                 "[#{window_offset_x}#,#{window_offset_y}] ,}"
+	                 "\"#{=21:pane_title}\" %H:%M %d-%b-%y"
 	},
 
 	{ .name = "status-right-length",
@@ -588,22 +580,6 @@ const struct options_table_entry options_table[] = {
 	  .default_num = 1
 	},
 
-	{ .name = "force-height",
-	  .type = OPTIONS_TABLE_NUMBER,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .minimum = 0,
-	  .maximum = INT_MAX,
-	  .default_num = 0
-	},
-
-	{ .name = "force-width",
-	  .type = OPTIONS_TABLE_NUMBER,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .minimum = 0,
-	  .maximum = INT_MAX,
-	  .default_num = 0
-	},
-
 	{ .name = "main-pane-height",
 	  .type = OPTIONS_TABLE_NUMBER,
 	  .scope = OPTIONS_TABLE_WINDOW,
@@ -620,27 +596,6 @@ const struct options_table_entry options_table[] = {
 	  .default_num = 80
 	},
 
-	{ .name = "mode-attr",
-	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 0,
-	  .style = "mode-style"
-	},
-
-	{ .name = "mode-bg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 3,
-	  .style = "mode-style"
-	},
-
-	{ .name = "mode-fg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 0,
-	  .style = "mode-style"
-	},
-
 	{ .name = "mode-keys",
 	  .type = OPTIONS_TABLE_CHOICE,
 	  .scope = OPTIONS_TABLE_WINDOW,
@@ -690,20 +645,6 @@ const struct options_table_entry options_table[] = {
 	  .default_num = 0
 	},
 
-	{ .name = "pane-active-border-bg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 8,
-	  .style = "pane-active-border-style"
-	},
-
-	{ .name = "pane-active-border-fg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 2,
-	  .style = "pane-active-border-style"
-	},
-
 	{ .name = "pane-active-border-style",
 	  .type = OPTIONS_TABLE_STYLE,
 	  .scope = OPTIONS_TABLE_WINDOW,
@@ -718,20 +659,6 @@ const struct options_table_entry options_table[] = {
 	  .default_num = 0
 	},
 
-	{ .name = "pane-border-bg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 8,
-	  .style = "pane-border-style"
-	},
-
-	{ .name = "pane-border-fg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 8,
-	  .style = "pane-border-style"
-	},
-
 	{ .name = "pane-border-format",
 	  .type = OPTIONS_TABLE_STRING,
 	  .scope = OPTIONS_TABLE_WINDOW,
@@ -770,101 +697,31 @@ const struct options_table_entry options_table[] = {
 	  .default_str = "default"
 	},
 
+	{ .name = "window-size",
+	  .type = OPTIONS_TABLE_CHOICE,
+	  .scope = OPTIONS_TABLE_WINDOW,
+	  .choices = options_table_window_size_list,
+	  .default_num = WINDOW_SIZE_SMALLEST
+	},
+
 	{ .name = "window-style",
 	  .type = OPTIONS_TABLE_STYLE,
 	  .scope = OPTIONS_TABLE_WINDOW,
 	  .default_str = "default"
 	},
 
-	{ .name = "window-status-activity-attr",
-	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = GRID_ATTR_REVERSE,
-	  .style = "window-status-activity-style"
-	},
-
-	{ .name = "window-status-activity-bg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 8,
-	  .style = "window-status-activity-style"
-	},
-
-	{ .name = "window-status-activity-fg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 8,
-	  .style = "window-status-activity-style"
-	},
-
 	{ .name = "window-status-activity-style",
 	  .type = OPTIONS_TABLE_STYLE,
 	  .scope = OPTIONS_TABLE_WINDOW,
 	  .default_str = "reverse"
 	},
 
-	{ .name = "window-status-attr",
-	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 0,
-	  .style = "window-status-style"
-	},
-
-	{ .name = "window-status-bell-attr",
-	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = GRID_ATTR_REVERSE,
-	  .style = "window-status-bell-style"
-	},
-
-	{ .name = "window-status-bell-bg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 8,
-	  .style = "window-status-bell-style"
-	},
-
-	{ .name = "window-status-bell-fg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 8,
-	  .style = "window-status-bell-style"
-	},
-
 	{ .name = "window-status-bell-style",
 	  .type = OPTIONS_TABLE_STYLE,
 	  .scope = OPTIONS_TABLE_WINDOW,
 	  .default_str = "reverse"
 	},
 
-	{ .name = "window-status-bg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 8,
-	  .style = "window-status-style"
-	},
-
-	{ .name = "window-status-current-attr",
-	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 0,
-	  .style = "window-status-current-style"
-	},
-
-	{ .name = "window-status-current-bg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 8,
-	  .style = "window-status-current-style"
-	},
-
-	{ .name = "window-status-current-fg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 8,
-	  .style = "window-status-current-style"
-	},
-
 	{ .name = "window-status-current-format",
 	  .type = OPTIONS_TABLE_STRING,
 	  .scope = OPTIONS_TABLE_WINDOW,
@@ -877,40 +734,12 @@ const struct options_table_entry options_table[] = {
 	  .default_str = "default"
 	},
 
-	{ .name = "window-status-fg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 8,
-	  .style = "window-status-style"
-	},
-
 	{ .name = "window-status-format",
 	  .type = OPTIONS_TABLE_STRING,
 	  .scope = OPTIONS_TABLE_WINDOW,
 	  .default_str = "#I:#W#{?window_flags,#{window_flags}, }"
 	},
 
-	{ .name = "window-status-last-attr",
-	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 0,
-	  .style = "window-status-last-style"
-	},
-
-	{ .name = "window-status-last-bg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 8,
-	  .style = "window-status-last-style"
-	},
-
-	{ .name = "window-status-last-fg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .scope = OPTIONS_TABLE_WINDOW,
-	  .default_num = 8,
-	  .style = "window-status-last-style"
-	},
-
 	{ .name = "window-status-last-style",
 	  .type = OPTIONS_TABLE_STYLE,
 	  .scope = OPTIONS_TABLE_WINDOW,
