@@ -1,7 +1,7 @@
 /* $OpenBSD$ */
 
 /*
- * Copyright (c) 2009 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2009 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -23,14 +23,17 @@
 
 #include <errno.h>
 #include <event.h>
+#include <limits.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 
+#include "tmux.h"
+
 #define is_runnable(p) \
-        ((p)->p_stat == LSRUN || (p)->p_stat == SIDL)
+	((p)->p_stat == LSRUN || (p)->p_stat == SIDL)
 #define is_stopped(p) \
-        ((p)->p_stat == SSTOP || (p)->p_stat == SZOMB)
+	((p)->p_stat == SSTOP || (p)->p_stat == SZOMB)
 
 struct kinfo_proc2	*cmp_procs(struct kinfo_proc2 *, struct kinfo_proc2 *);
 char			*osdep_get_name(int, char *);
@@ -80,30 +83,32 @@ osdep_get_name(int fd, __unused char *tty)
 		return (NULL);
 
 	buf = NULL;
-	len = sizeof(bestp);
+	len = sizeof bestp;
+
 	mib[0] = CTL_KERN;
 	mib[1] = KERN_PROC2;
 	mib[2] = KERN_PROC_PGRP;
-	mib[4] = sizeof (*buf);
-	mib[5] = 0;
+	mib[4] = sizeof *buf;
 
 retry:
+	mib[5] = 0;
+
 	if (sysctl(mib, __arraycount(mib), NULL, &len, NULL, 0) == -1)
 		return (NULL);
 
-	if ((newbuf = realloc(buf, len * sizeof (*buf))) == NULL)
+	if ((newbuf = realloc(buf, len)) == NULL)
 		goto error;
 	buf = newbuf;
 
-	mib[5] = len / sizeof(*buf);
+	mib[5] = len / (sizeof *buf);
 	if (sysctl(mib, __arraycount(mib), buf, &len, NULL, 0) == -1) {
 		if (errno == ENOMEM)
-			goto retry; /* possible infinite loop? */
+			goto retry;
 		goto error;
 	}
 
 	bestp = NULL;
-	for (i = 0; i < len / sizeof (*buf); i++) {
+	for (i = 0; i < len / (sizeof *buf); i++) {
 		if (buf[i].p_tdev != sb.st_rdev)
 			continue;
 		if (bestp == NULL)
@@ -127,6 +132,22 @@ error:
 char *
 osdep_get_cwd(int fd)
 {
+	static char	target[PATH_MAX + 1];
+	char		*path;
+	pid_t		pgrp;
+	ssize_t		n;
+
+	if ((pgrp = tcgetpgrp(fd)) == -1)
+		return (NULL);
+
+	xasprintf(&path, "/proc/%lld/cwd", (long long) pgrp);
+	n = readlink(path, target, sizeof(target) - 1);
+	free(path);
+	if (n > 0) {
+		target[n] = '\0';
+		return (target);
+	}
+
 	return (NULL);
 }
 
