@@ -54,12 +54,53 @@ struct screen_write_collect_line {
 	TAILQ_HEAD(, screen_write_collect_item) items;
 };
 
+static void
+screen_write_offset_timer(__unused int fd, __unused short events, void *data)
+{
+	struct window	*w = data;
+
+	tty_update_window_offset(w);
+}
+
+/* Set cursor position. */
+static void
+screen_write_set_cursor(struct screen_write_ctx *ctx, int cx, int cy)
+{
+	struct window_pane	*wp = ctx->wp;
+	struct window		*w;
+	struct screen		*s = ctx->s;
+	struct timeval		 tv = { .tv_usec = 10000 };
+
+	if (cx != -1 && (u_int)cx == s->cx && cy != -1 && (u_int)cy == s->cy)
+		return;
+
+	if (cx != -1) {
+		if ((u_int)cx > screen_size_x(s)) /* allow last column */
+			cx = screen_size_x(s) - 1;
+		s->cx = cx;
+	}
+	if (cy != -1) {
+		if ((u_int)cy > screen_size_y(s) - 1)
+			cy = screen_size_y(s) - 1;
+		s->cy = cy;
+	}
+
+	if (wp == NULL)
+		return;
+	w = wp->window;
+
+	if (!event_initialized(&w->offset_timer))
+		evtimer_set(&w->offset_timer, screen_write_offset_timer, w);
+	if (!evtimer_pending(&w->offset_timer, NULL))
+		evtimer_add(&w->offset_timer, &tv);
+}
+
 /* Initialize writing with a window. */
 void
 screen_write_start(struct screen_write_ctx *ctx, struct window_pane *wp,
     struct screen *s)
 {
-	char	tmp[16];
+	char	tmp[32];
 	u_int	y;
 
 	memset(ctx, 0, sizeof *ctx);
@@ -78,8 +119,10 @@ screen_write_start(struct screen_write_ctx *ctx, struct window_pane *wp,
 	ctx->scrolled = 0;
 	ctx->bg = 8;
 
-	if (wp != NULL)
-		snprintf(tmp, sizeof tmp, "pane %%%u", wp->id);
+	if (wp != NULL) {
+		snprintf(tmp, sizeof tmp, "pane %%%u (at %u,%u)", wp->id,
+		    wp->xoff, wp->yoff);
+	}
 	log_debug("%s: size %ux%u, %s", __func__, screen_size_x(ctx->s),
 	    screen_size_y(ctx->s), wp == NULL ? "no pane" : tmp);
 }
@@ -107,11 +150,10 @@ screen_write_reset(struct screen_write_ctx *ctx)
 	screen_reset_tabs(s);
 	screen_write_scrollregion(ctx, 0, screen_size_y(s) - 1);
 
-	s->mode &= ~(MODE_INSERT|MODE_KCURSOR|MODE_KKEYPAD|MODE_FOCUSON);
-	s->mode &= ~(ALL_MOUSE_MODES|MODE_MOUSE_UTF8|MODE_MOUSE_SGR);
+	s->mode = MODE_CURSOR | MODE_WRAP;
 
 	screen_write_clearscreen(ctx, 8);
-	screen_write_cursormove(ctx, 0, 0);
+	screen_write_set_cursor(ctx, 0, 0);
 }
 
 /* Write character. */
@@ -127,41 +169,6 @@ screen_write_putc(struct screen_write_ctx *ctx, const struct grid_cell *gcp,
 	screen_write_cell(ctx, &gc);
 }
 
-/* Calculate string length, with embedded formatting. */
-size_t
-screen_write_cstrlen(const char *fmt, ...)
-{
-	va_list	ap;
-	char   *msg, *msg2, *ptr, *ptr2;
-	size_t	size;
-
-	va_start(ap, fmt);
-	xvasprintf(&msg, fmt, ap);
-	va_end(ap);
-	msg2 = xmalloc(strlen(msg) + 1);
-
-	ptr = msg;
-	ptr2 = msg2;
-	while (*ptr != '\0') {
-		if (ptr[0] == '#' && ptr[1] == '[') {
-			while (*ptr != ']' && *ptr != '\0')
-				ptr++;
-			if (*ptr == ']')
-				ptr++;
-			continue;
-		}
-		*ptr2++ = *ptr++;
-	}
-	*ptr2 = '\0';
-
-	size = screen_write_strlen("%s", msg2);
-
-	free(msg);
-	free(msg2);
-
-	return (size);
-}
-
 /* Calculate string length. */
 size_t
 screen_write_strlen(const char *fmt, ...)
@@ -280,77 +287,6 @@ screen_write_vnputs(struct screen_write_ctx *ctx, ssize_t maxlen,
 	free(msg);
 }
 
-/* Write string, similar to nputs, but with embedded formatting (#[]). */
-void
-screen_write_cnputs(struct screen_write_ctx *ctx, ssize_t maxlen,
-    const struct grid_cell *gcp, const char *fmt, ...)
-{
-	struct grid_cell	 gc;
-	struct utf8_data	*ud = &gc.data;
-	va_list			 ap;
-	char			*msg;
-	u_char 			*ptr, *last;
-	size_t			 left, size = 0;
-	enum utf8_state		 more;
-
-	memcpy(&gc, gcp, sizeof gc);
-
-	va_start(ap, fmt);
-	xvasprintf(&msg, fmt, ap);
-	va_end(ap);
-
-	ptr = msg;
-	while (*ptr != '\0') {
-		if (ptr[0] == '#' && ptr[1] == '[') {
-			ptr += 2;
-			last = ptr + strcspn(ptr, "]");
-			if (*last == '\0') {
-				/* No ]. Not much point in doing anything. */
-				break;
-			}
-			*last = '\0';
-
-			style_parse(gcp, &gc, ptr);
-			ptr = last + 1;
-			continue;
-		}
-
-		if (*ptr > 0x7f && utf8_open(ud, *ptr) == UTF8_MORE) {
-			ptr++;
-
-			left = strlen(ptr);
-			if (left < (size_t)ud->size - 1)
-				break;
-			while ((more = utf8_append(ud, *ptr)) == UTF8_MORE)
-				ptr++;
-			ptr++;
-
-			if (more != UTF8_DONE)
-				continue;
-			if (maxlen > 0 && size + ud->width > (size_t)maxlen) {
-				while (size < (size_t)maxlen) {
-					screen_write_putc(ctx, &gc, ' ');
-					size++;
-				}
-				break;
-			}
-			size += ud->width;
-			screen_write_cell(ctx, &gc);
-		} else {
-			if (maxlen > 0 && size + 1 > (size_t)maxlen)
-				break;
-
-			if (*ptr > 0x1f && *ptr < 0x7f) {
-				size++;
-				screen_write_putc(ctx, &gc, *ptr);
-			}
-			ptr++;
-		}
-	}
-
-	free(msg);
-}
-
 /* Copy from another screen. Assumes target region is big enough. */
 void
 screen_write_copy(struct screen_write_ctx *ctx, struct screen *src, u_int px,
@@ -382,13 +318,13 @@ screen_write_copy(struct screen_write_ctx *ctx, struct screen *src, u_int px,
 				screen_write_cell(ctx, &gc);
 		}
 		cy++;
-		screen_write_cursormove(ctx, cx, cy);
+		screen_write_set_cursor(ctx, cx, cy);
 	}
 }
 
 /*
  * Copy from another screen but without the selection stuff. Also assumes the
- * target region is already big enough and already cleared.
+ * target region is already big enough.
  */
 void
 screen_write_fast_copy(struct screen_write_ctx *ctx, struct screen *src,
@@ -413,8 +349,7 @@ screen_write_fast_copy(struct screen_write_ctx *ctx, struct screen *src,
 			grid_get_cell(gd, xx, yy, &gc);
 			if (xx + gc.data.width > px + nx)
 				break;
-			if (!grid_cells_equal(&gc, &grid_default_cell))
-				grid_view_set_cell(ctx->s->grid, cx, cy, &gc);
+			grid_view_set_cell(ctx->s->grid, cx, cy, &gc);
 			cx++;
 		}
 		cy++;
@@ -440,7 +375,7 @@ screen_write_hline(struct screen_write_ctx *ctx, u_int nx, int left, int right)
 		screen_write_putc(ctx, &gc, 'q');
 	screen_write_putc(ctx, &gc, right ? 'u' : 'q');
 
-	screen_write_cursormove(ctx, cx, cy);
+	screen_write_set_cursor(ctx, cx, cy);
 }
 
 /* Draw a horizontal line on screen. */
@@ -459,13 +394,13 @@ screen_write_vline(struct screen_write_ctx *ctx, u_int ny, int top, int bottom)
 
 	screen_write_putc(ctx, &gc, top ? 'w' : 'x');
 	for (i = 1; i < ny - 1; i++) {
-		screen_write_cursormove(ctx, cx, cy + i);
+		screen_write_set_cursor(ctx, cx, cy + i);
 		screen_write_putc(ctx, &gc, 'x');
 	}
-	screen_write_cursormove(ctx, cx, cy + ny - 1);
+	screen_write_set_cursor(ctx, cx, cy + ny - 1);
 	screen_write_putc(ctx, &gc, bottom ? 'v' : 'x');
 
-	screen_write_cursormove(ctx, cx, cy);
+	screen_write_set_cursor(ctx, cx, cy);
 }
 
 /* Draw a box on screen. */
@@ -487,22 +422,22 @@ screen_write_box(struct screen_write_ctx *ctx, u_int nx, u_int ny)
 		screen_write_putc(ctx, &gc, 'q');
 	screen_write_putc(ctx, &gc, 'k');
 
-	screen_write_cursormove(ctx, cx, cy + ny - 1);
+	screen_write_set_cursor(ctx, cx, cy + ny - 1);
 	screen_write_putc(ctx, &gc, 'm');
 	for (i = 1; i < nx - 1; i++)
 		screen_write_putc(ctx, &gc, 'q');
 	screen_write_putc(ctx, &gc, 'j');
 
 	for (i = 1; i < ny - 1; i++) {
-		screen_write_cursormove(ctx, cx, cy + i);
+		screen_write_set_cursor(ctx, cx, cy + i);
 		screen_write_putc(ctx, &gc, 'x');
 	}
 	for (i = 1; i < ny - 1; i++) {
-		screen_write_cursormove(ctx, cx + nx - 1, cy + i);
+		screen_write_set_cursor(ctx, cx + nx - 1, cy + i);
 		screen_write_putc(ctx, &gc, 'x');
 	}
 
-	screen_write_cursormove(ctx, cx, cy);
+	screen_write_set_cursor(ctx, cx, cy);
 }
 
 /*
@@ -557,7 +492,7 @@ screen_write_preview(struct screen_write_ctx *ctx, struct screen *src, u_int nx,
 	if (src->mode & MODE_CURSOR) {
 		grid_view_get_cell(src->grid, src->cx, src->cy, &gc);
 		gc.attr |= GRID_ATTR_REVERSE;
-		screen_write_cursormove(ctx, cx + (src->cx - px),
+		screen_write_set_cursor(ctx, cx + (src->cx - px),
 		    cy + (src->cy - py));
 		screen_write_cell(ctx, &gc);
 	}
@@ -603,25 +538,26 @@ void
 screen_write_cursorup(struct screen_write_ctx *ctx, u_int ny)
 {
 	struct screen	*s = ctx->s;
+	u_int		 cx = s->cx, cy = s->cy;
 
 	if (ny == 0)
 		ny = 1;
 
-	if (s->cy < s->rupper) {
+	if (cy < s->rupper) {
 		/* Above region. */
-		if (ny > s->cy)
-			ny = s->cy;
+		if (ny > cy)
+			ny = cy;
 	} else {
 		/* Below region. */
-		if (ny > s->cy - s->rupper)
-			ny = s->cy - s->rupper;
+		if (ny > cy - s->rupper)
+			ny = cy - s->rupper;
 	}
-	if (s->cx == screen_size_x(s))
-	    s->cx--;
-	if (ny == 0)
-		return;
+	if (cx == screen_size_x(s))
+		cx--;
 
-	s->cy -= ny;
+	cy -= ny;
+
+	screen_write_set_cursor(ctx, cx, cy);
 }
 
 /* Cursor down by ny. */
@@ -629,25 +565,28 @@ void
 screen_write_cursordown(struct screen_write_ctx *ctx, u_int ny)
 {
 	struct screen	*s = ctx->s;
+	u_int		 cx = s->cx, cy = s->cy;
 
 	if (ny == 0)
 		ny = 1;
 
-	if (s->cy > s->rlower) {
+	if (cy > s->rlower) {
 		/* Below region. */
-		if (ny > screen_size_y(s) - 1 - s->cy)
-			ny = screen_size_y(s) - 1 - s->cy;
+		if (ny > screen_size_y(s) - 1 - cy)
+			ny = screen_size_y(s) - 1 - cy;
 	} else {
 		/* Above region. */
-		if (ny > s->rlower - s->cy)
-			ny = s->rlower - s->cy;
+		if (ny > s->rlower - cy)
+			ny = s->rlower - cy;
 	}
-	if (s->cx == screen_size_x(s))
-	    s->cx--;
-	if (ny == 0)
+	if (cx == screen_size_x(s))
+	    cx--;
+	else if (ny == 0)
 		return;
 
-	s->cy += ny;
+	cy += ny;
+
+	screen_write_set_cursor(ctx, cx, cy);
 }
 
 /* Cursor right by nx. */
@@ -655,16 +594,19 @@ void
 screen_write_cursorright(struct screen_write_ctx *ctx, u_int nx)
 {
 	struct screen	*s = ctx->s;
+	u_int		 cx = s->cx, cy = s->cy;
 
 	if (nx == 0)
 		nx = 1;
 
-	if (nx > screen_size_x(s) - 1 - s->cx)
-		nx = screen_size_x(s) - 1 - s->cx;
+	if (nx > screen_size_x(s) - 1 - cx)
+		nx = screen_size_x(s) - 1 - cx;
 	if (nx == 0)
 		return;
 
-	s->cx += nx;
+	cx += nx;
+
+	screen_write_set_cursor(ctx, cx, cy);
 }
 
 /* Cursor left by nx. */
@@ -672,16 +614,19 @@ void
 screen_write_cursorleft(struct screen_write_ctx *ctx, u_int nx)
 {
 	struct screen	*s = ctx->s;
+	u_int		 cx = s->cx, cy = s->cy;
 
 	if (nx == 0)
 		nx = 1;
 
-	if (nx > s->cx)
-		nx = s->cx;
+	if (nx > cx)
+		nx = cx;
 	if (nx == 0)
 		return;
 
-	s->cx -= nx;
+	cx -= nx;
+
+	screen_write_set_cursor(ctx, cx, cy);
 }
 
 /* Backspace; cursor left unless at start of wrapped line when can move up. */
@@ -690,17 +635,20 @@ screen_write_backspace(struct screen_write_ctx *ctx)
 {
 	struct screen		*s = ctx->s;
 	struct grid_line	*gl;
+	u_int			 cx = s->cx, cy = s->cy;
 
-	if (s->cx == 0) {
-		if (s->cy == 0)
+	if (cx == 0) {
+		if (cy == 0)
 			return;
-		gl = grid_get_line(s->grid, s->grid->hsize + s->cy - 1);
+		gl = grid_get_line(s->grid, s->grid->hsize + cy - 1);
 		if (gl->flags & GRID_LINE_WRAPPED) {
-			s->cy--;
-			s->cx = screen_size_x(s) - 1;
+			cy--;
+			cx = screen_size_x(s) - 1;
 		}
 	} else
-		s->cx--;
+		cx--;
+
+	screen_write_set_cursor(ctx, cx, cy);
 }
 
 /* VT100 alignment test. */
@@ -712,8 +660,6 @@ screen_write_alignmenttest(struct screen_write_ctx *ctx)
 	struct grid_cell       	 gc;
 	u_int			 xx, yy;
 
-	screen_write_initctx(ctx, &ttyctx);
-
 	memcpy(&gc, &grid_default_cell, sizeof gc);
 	utf8_set(&gc.data, 'E');
 
@@ -722,12 +668,13 @@ screen_write_alignmenttest(struct screen_write_ctx *ctx)
 			grid_view_set_cell(s->grid, xx, yy, &gc);
 	}
 
-	s->cx = 0;
-	s->cy = 0;
+	screen_write_set_cursor(ctx, 0, 0);
 
 	s->rupper = 0;
 	s->rlower = screen_size_y(s) - 1;
 
+	screen_write_initctx(ctx, &ttyctx);
+
 	screen_write_collect_clear(ctx, 0, screen_size_y(s) - 1);
 	tty_write(tty_cmd_alignmenttest, &ttyctx);
 }
@@ -918,7 +865,7 @@ screen_write_clearline(struct screen_write_ctx *ctx, u_int bg)
 	u_int			 sx = screen_size_x(s);
 
 	gl = grid_get_line(s->grid, s->grid->hsize + s->cy);
-	if (gl->cellsize == 0 && bg == 8)
+	if (gl->cellsize == 0 && COLOUR_DEFAULT(bg))
 		return;
 
 	screen_write_initctx(ctx, &ttyctx);
@@ -941,7 +888,7 @@ screen_write_clearendofline(struct screen_write_ctx *ctx, u_int bg)
 	u_int			 sx = screen_size_x(s);
 
 	gl = grid_get_line(s->grid, s->grid->hsize + s->cy);
-	if (s->cx > sx - 1 || (s->cx >= gl->cellsize && bg == 8))
+	if (s->cx > sx - 1 || (s->cx >= gl->cellsize && COLOUR_DEFAULT(bg)))
 		return;
 
 	screen_write_initctx(ctx, &ttyctx);
@@ -979,17 +926,24 @@ screen_write_clearstartofline(struct screen_write_ctx *ctx, u_int bg)
 
 /* Move cursor to px,py. */
 void
-screen_write_cursormove(struct screen_write_ctx *ctx, u_int px, u_int py)
+screen_write_cursormove(struct screen_write_ctx *ctx, int px, int py,
+    int origin)
 {
 	struct screen	*s = ctx->s;
 
-	if (px > screen_size_x(s) - 1)
+	if (origin && py != -1 && (s->mode & MODE_ORIGIN)) {
+		if ((u_int)py > s->rlower - s->rupper)
+			py = s->rlower;
+		else
+			py += s->rupper;
+	}
+
+	if (px != -1 && (u_int)px > screen_size_x(s) - 1)
 		px = screen_size_x(s) - 1;
-	if (py > screen_size_y(s) - 1)
+	if (py != -1 && (u_int)py > screen_size_y(s) - 1)
 		py = screen_size_y(s) - 1;
 
-	s->cx = px;
-	s->cy = py;
+	screen_write_set_cursor(ctx, px, py);
 }
 
 /* Reverse index (up with scroll). */
@@ -1005,7 +959,7 @@ screen_write_reverseindex(struct screen_write_ctx *ctx, u_int bg)
 	if (s->cy == s->rupper)
 		grid_view_scroll_region_down(s->grid, s->rupper, s->rlower, bg);
 	else if (s->cy > 0)
-		s->cy--;
+		screen_write_set_cursor(ctx, -1, s->cy - 1);
 
 	screen_write_collect_flush(ctx, 0);
 	tty_write(tty_cmd_reverseindex, &ttyctx);
@@ -1028,8 +982,7 @@ screen_write_scrollregion(struct screen_write_ctx *ctx, u_int rupper,
 	screen_write_collect_flush(ctx, 0);
 
 	/* Cursor moves to top-left. */
-	s->cx = 0;
-	s->cy = 0;
+	screen_write_set_cursor(ctx, 0, 0);
 
 	s->rupper = rupper;
 	s->rlower = rlower;
@@ -1062,7 +1015,7 @@ screen_write_linefeed(struct screen_write_ctx *ctx, int wrapped, u_int bg)
 		screen_write_collect_scroll(ctx);
 		ctx->scrolled++;
 	} else if (s->cy < screen_size_y(s) - 1)
-		s->cy++;
+		screen_write_set_cursor(ctx, -1, s->cy + 1);
 }
 
 /* Scroll up. */
@@ -1094,9 +1047,7 @@ screen_write_scrollup(struct screen_write_ctx *ctx, u_int lines, u_int bg)
 void
 screen_write_carriagereturn(struct screen_write_ctx *ctx)
 {
-	struct screen	*s = ctx->s;
-
-	s->cx = 0;
+	screen_write_set_cursor(ctx, 0, -1);
 }
 
 /* Clear to end of screen from cursor. */
@@ -1251,7 +1202,7 @@ screen_write_collect_flush(struct screen_write_ctx *ctx, int scroll_only)
 	cx = s->cx; cy = s->cy;
 	for (y = 0; y < screen_size_y(s); y++) {
 		TAILQ_FOREACH_SAFE(ci, &ctx->list[y].items, entry, tmp) {
-			screen_write_cursormove(ctx, ci->x, y);
+			screen_write_set_cursor(ctx, ci->x, y);
 			screen_write_initctx(ctx, &ttyctx);
 			ttyctx.cell = &ci->gc;
 			ttyctx.wrapped = ci->wrapped;
@@ -1300,14 +1251,15 @@ screen_write_collect_end(struct screen_write_ctx *ctx)
 			grid_view_set_cell(s->grid, xx, s->cy,
 			    &grid_default_cell);
 		}
-		if (gc.data.width > 1)
+		if (gc.data.width > 1) {
 			grid_view_set_cell(s->grid, xx, s->cy,
 			    &grid_default_cell);
+		}
 	}
 
 	memcpy(&gc, &ci->gc, sizeof gc);
 	grid_view_set_cells(s->grid, s->cx, s->cy, &gc, ci->data, ci->used);
-	s->cx += ci->used;
+	screen_write_set_cursor(ctx, s->cx + ci->used, -1);
 
 	for (xx = s->cx; xx < screen_size_x(s); xx++) {
 		grid_view_get_cell(s->grid, xx, s->cy, &gc);
@@ -1361,7 +1313,7 @@ screen_write_collect_add(struct screen_write_ctx *ctx,
 		log_debug("%s: wrapped at %u,%u", __func__, s->cx, s->cy);
 		ci->wrapped = 1;
 		screen_write_linefeed(ctx, 1, 8);
-		s->cx = 0;
+		screen_write_set_cursor(ctx, 0, -1);
 	}
 
 	if (ci->used == 0)
@@ -1395,7 +1347,7 @@ screen_write_cell(struct screen_write_ctx *ctx, const struct grid_cell *gc)
 		screen_write_collect_flush(ctx, 0);
 		if ((gc = screen_write_combine(ctx, &gc->data, &xx)) != 0) {
 			cx = s->cx; cy = s->cy;
-			screen_write_cursormove(ctx, xx, s->cy);
+			screen_write_set_cursor(ctx, xx, s->cy);
 			screen_write_initctx(ctx, &ttyctx);
 			ttyctx.cell = gc;
 			tty_write(tty_cmd_cell, &ttyctx);
@@ -1423,7 +1375,7 @@ screen_write_cell(struct screen_write_ctx *ctx, const struct grid_cell *gc)
 	if ((s->mode & MODE_WRAP) && s->cx > sx - width) {
 		log_debug("%s: wrapped at %u,%u", __func__, s->cx, s->cy);
 		screen_write_linefeed(ctx, 1, 8);
-		s->cx = 0;
+		screen_write_set_cursor(ctx, 0, -1);
 		screen_write_collect_flush(ctx, 1);
 	}
 
@@ -1496,9 +1448,9 @@ screen_write_cell(struct screen_write_ctx *ctx, const struct grid_cell *gc)
 	 */
 	last = !(s->mode & MODE_WRAP);
 	if (s->cx <= sx - last - width)
-		s->cx += width;
+		screen_write_set_cursor(ctx, s->cx + width, -1);
 	else
-		s->cx = sx - last;
+		screen_write_set_cursor(ctx,  sx - last, -1);
 
 	/* Create space for character in insert mode. */
 	if (s->mode & MODE_INSERT) {
