@@ -29,14 +29,6 @@
 
 #include "tmux.h"
 
-static char	*status_redraw_get_left(struct client *, time_t,
-		     struct grid_cell *, size_t *);
-static char	*status_redraw_get_right(struct client *, time_t,
-		     struct grid_cell *, size_t *);
-static char	*status_print(struct client *, struct winlink *, time_t,
-		     struct grid_cell *);
-static char	*status_replace(struct client *, struct winlink *, const char *,
-		     time_t);
 static void	 status_message_callback(int, short, void *);
 static void	 status_timer_callback(int, short, void *);
 
@@ -45,8 +37,8 @@ static const char *status_prompt_up_history(u_int *);
 static const char *status_prompt_down_history(u_int *);
 static void	 status_prompt_add_history(const char *);
 
-static const char **status_prompt_complete_list(u_int *, const char *);
-static char	*status_prompt_complete_prefix(const char **, u_int);
+static char    **status_prompt_complete_list(u_int *, const char *);
+static char	*status_prompt_complete_prefix(char **, u_int);
 static char	*status_prompt_complete(struct session *, const char *);
 
 /* Status prompt history. */
@@ -157,7 +149,7 @@ status_timer_callback(__unused int fd, __unused short events, void *arg)
 		return;
 
 	if (c->message_string == NULL && c->prompt_string == NULL)
-		c->flags |= CLIENT_STATUS;
+		c->flags |= CLIENT_REDRAWSTATUS;
 
 	timerclear(&tv);
 	tv.tv_sec = options_get_number(s->options, "status-interval");
@@ -194,9 +186,10 @@ status_timer_start_all(void)
 
 /* Update status cache. */
 void
-status_update_saved(struct session *s)
+status_update_cache(struct session *s)
 {
-	if (!options_get_number(s->options, "status"))
+	s->statuslines = options_get_number(s->options, "status");
+	if (s->statuslines == 0)
 		s->statusat = -1;
 	else if (options_get_number(s->options, "status-position") == 0)
 		s->statusat = 0;
@@ -214,370 +207,201 @@ status_at_line(struct client *c)
 		return (-1);
 	if (s->statusat != 1)
 		return (s->statusat);
-	return (c->tty.sy - status_line_size(s));
+	return (c->tty.sy - status_line_size(c));
 }
 
-/*
- * Get size of status line for session. 0 means off. Note that status line may
- * be forced off for an individual client if it is too small (the
- * CLIENT_STATUSOFF flag is set for this).
- */
+/* Get size of status line for client's session. 0 means off. */
 u_int
-status_line_size(struct session *s)
+status_line_size(struct client *c)
 {
-	if (s->statusat == -1)
+	struct session	*s = c->session;
+
+	if (c->flags & CLIENT_STATUSOFF)
 		return (0);
-	return (1);
-}
-
-/* Retrieve options for left string. */
-static char *
-status_redraw_get_left(struct client *c, time_t t, struct grid_cell *gc,
-    size_t *size)
-{
-	struct session	*s = c->session;
-	const char	*template;
-	char		*left;
-	size_t		 leftlen;
-
-	style_apply_update(gc, s->options, "status-left-style");
-
-	template = options_get_string(s->options, "status-left");
-	left = status_replace(c, NULL, template, t);
-
-	*size = options_get_number(s->options, "status-left-length");
-	leftlen = screen_write_cstrlen("%s", left);
-	if (leftlen < *size)
-		*size = leftlen;
-	return (left);
-}
-
-/* Retrieve options for right string. */
-static char *
-status_redraw_get_right(struct client *c, time_t t, struct grid_cell *gc,
-    size_t *size)
-{
-	struct session	*s = c->session;
-	const char	*template;
-	char		*right;
-	size_t		 rightlen;
-
-	style_apply_update(gc, s->options, "status-right-style");
-
-	template = options_get_string(s->options, "status-right");
-	right = status_replace(c, NULL, template, t);
-
-	*size = options_get_number(s->options, "status-right-length");
-	rightlen = screen_write_cstrlen("%s", right);
-	if (rightlen < *size)
-		*size = rightlen;
-	return (right);
+	return (s->statuslines);
 }
 
 /* Get window at window list position. */
-struct window *
-status_get_window_at(struct client *c, u_int x)
+struct style_range *
+status_get_range(struct client *c, u_int x, u_int y)
 {
-	struct session	*s = c->session;
-	struct winlink	*wl;
-	struct options	*oo;
-	const char	*sep;
-	size_t		 seplen;
+	struct status_line	*sl = &c->status;
+	struct style_range	*sr;
 
-	x += c->wlmouse;
-	RB_FOREACH(wl, winlinks, &s->windows) {
-		oo = wl->window->options;
-
-		sep = options_get_string(oo, "window-status-separator");
-		seplen = screen_write_cstrlen("%s", sep);
-
-		if (x < wl->status_width)
-			return (wl->window);
-		x -= wl->status_width + seplen;
+	if (y >= nitems(sl->entries))
+		return (NULL);
+	TAILQ_FOREACH(sr, &sl->entries[y].ranges, entry) {
+		if (x >= sr->start && x < sr->end)
+			return (sr);
 	}
 	return (NULL);
 }
 
-/* Draw status for client on the last lines of given context. */
+/* Free all ranges. */
+static void
+status_free_ranges(struct style_ranges *srs)
+{
+	struct style_range	*sr, *sr1;
+
+	TAILQ_FOREACH_SAFE(sr, srs, entry, sr1) {
+		TAILQ_REMOVE(srs, sr, entry);
+		free(sr);
+	}
+}
+
+/* Save old status line. */
+static void
+status_push_screen(struct client *c)
+{
+	struct status_line *sl = &c->status;
+
+	if (sl->active == &sl->screen) {
+		sl->active = xmalloc(sizeof *sl->active);
+		screen_init(sl->active, c->tty.sx, status_line_size(c), 0);
+	}
+	sl->references++;
+}
+
+/* Restore old status line. */
+static void
+status_pop_screen(struct client *c)
+{
+	struct status_line *sl = &c->status;
+
+	if (--sl->references == 0) {
+		screen_free(sl->active);
+		free(sl->active);
+		sl->active = &sl->screen;
+	}
+}
+
+/* Initialize status line. */
+void
+status_init(struct client *c)
+{
+	struct status_line	*sl = &c->status;
+	u_int			 i;
+
+	for (i = 0; i < nitems(sl->entries); i++)
+		TAILQ_INIT(&sl->entries[i].ranges);
+
+	screen_init(&sl->screen, c->tty.sx, 1, 0);
+	sl->active = &sl->screen;
+}
+
+/* Free status line. */
+void
+status_free(struct client *c)
+{
+	struct status_line	*sl = &c->status;
+	u_int			 i;
+
+	for (i = 0; i < nitems(sl->entries); i++) {
+		status_free_ranges(&sl->entries[i].ranges);
+		free((void *)sl->entries[i].expanded);
+	}
+
+	if (event_initialized(&sl->timer))
+		evtimer_del(&sl->timer);
+
+	if (sl->active != &sl->screen) {
+		screen_free(sl->active);
+		free(sl->active);
+	}
+	screen_free(&sl->screen);
+}
+
+/* Draw status line for client. */
 int
 status_redraw(struct client *c)
 {
-	struct screen_write_ctx	 ctx;
-	struct session		*s = c->session;
-	struct winlink		*wl;
-	struct screen		 old_status, window_list;
-	struct grid_cell	 stdgc, lgc, rgc, gc;
-	struct options		*oo;
-	time_t			 t;
-	char			*left, *right;
-	const char		*sep;
-	u_int			 offset, needed, lines;
-	u_int			 wlstart, wlwidth, wlavailable, wloffset, wlsize;
-	size_t			 llen, rlen, seplen;
-	int			 larrow, rarrow;
+	struct status_line		*sl = &c->status;
+	struct status_line_entry	*sle;
+	struct session			*s = c->session;
+	struct screen_write_ctx		 ctx;
+	struct grid_cell		 gc;
+	u_int				 lines, i, width = c->tty.sx;
+	int				 flags, force = 0, changed = 0;
+	struct options_entry		*o;
+	struct format_tree		*ft;
+	const char			*fmt;
+	char				*expanded;
 
-	/* Delete the saved status line, if any. */
-	if (c->status.old_status != NULL) {
-		screen_free(c->status.old_status);
-		free(c->status.old_status);
-		c->status.old_status = NULL;
-	}
+	log_debug("%s enter", __func__);
+
+	/* Shouldn't get here if not the active screen. */
+	if (sl->active != &sl->screen)
+		fatalx("not the active screen");
 
 	/* No status line? */
-	lines = status_line_size(s);
+	lines = status_line_size(c);
 	if (c->tty.sy == 0 || lines == 0)
 		return (1);
-	left = right = NULL;
-	larrow = rarrow = 0;
-
-	/* Store current time. */
-	t = time(NULL);
 
 	/* Set up default colour. */
-	style_apply(&stdgc, s->options, "status-style");
-
-	/* Create the target screen. */
-	memcpy(&old_status, &c->status.status, sizeof old_status);
-	screen_init(&c->status.status, c->tty.sx, lines, 0);
-	screen_write_start(&ctx, NULL, &c->status.status);
-	for (offset = 0; offset < lines * c->tty.sx; offset++)
-		screen_write_putc(&ctx, &stdgc, ' ');
-	screen_write_stop(&ctx);
-
-	/* If the height is too small, blank status line. */
-	if (c->tty.sy < lines)
-		goto out;
-
-	/* Work out left and right strings. */
-	memcpy(&lgc, &stdgc, sizeof lgc);
-	left = status_redraw_get_left(c, t, &lgc, &llen);
-	memcpy(&rgc, &stdgc, sizeof rgc);
-	right = status_redraw_get_right(c, t, &rgc, &rlen);
-
-	/*
-	 * Figure out how much space we have for the window list. If there
-	 * isn't enough space, just show a blank status line.
-	 */
-	needed = 0;
-	if (llen != 0)
-		needed += llen;
-	if (rlen != 0)
-		needed += rlen;
-	if (c->tty.sx == 0 || c->tty.sx <= needed)
-		goto out;
-	wlavailable = c->tty.sx - needed;
-
-	/* Calculate the total size needed for the window list. */
-	wlstart = wloffset = wlwidth = 0;
-	RB_FOREACH(wl, winlinks, &s->windows) {
-		free(wl->status_text);
-		memcpy(&wl->status_cell, &stdgc, sizeof wl->status_cell);
-		wl->status_text = status_print(c, wl, t, &wl->status_cell);
-		wl->status_width = screen_write_cstrlen("%s", wl->status_text);
-
-		if (wl == s->curw)
-			wloffset = wlwidth;
-
-		oo = wl->window->options;
-		sep = options_get_string(oo, "window-status-separator");
-		seplen = screen_write_cstrlen("%s", sep);
-		wlwidth += wl->status_width + seplen;
+	style_apply(&gc, s->options, "status-style");
+	if (!grid_cells_equal(&gc, &sl->style)) {
+		force = 1;
+		memcpy(&sl->style, &gc, sizeof sl->style);
 	}
 
-	/* Create a new screen for the window list. */
-	screen_init(&window_list, wlwidth, 1, 0);
-
-	/* And draw the window list into it. */
-	screen_write_start(&ctx, NULL, &window_list);
-	RB_FOREACH(wl, winlinks, &s->windows) {
-		screen_write_cnputs(&ctx, -1, &wl->status_cell, "%s",
-		    wl->status_text);
-
-		oo = wl->window->options;
-		sep = options_get_string(oo, "window-status-separator");
-		screen_write_cnputs(&ctx, -1, &stdgc, "%s", sep);
+	/* Resize the target screen. */
+	if (screen_size_x(&sl->screen) != width ||
+	    screen_size_y(&sl->screen) != lines) {
+		if (screen_size_x(&sl->screen) != width)
+			force = 1;
+		screen_resize(&sl->screen, width, lines, 0);
+		changed = 1;
 	}
-	screen_write_stop(&ctx);
+	screen_write_start(&ctx, NULL, &sl->screen);
 
-	/* If there is enough space for the total width, skip to draw now. */
-	if (wlwidth <= wlavailable)
-		goto draw;
-
-	/* Find size of current window text. */
-	wlsize = s->curw->status_width;
-
-	/*
-	 * If the current window is already on screen, good to draw from the
-	 * start and just leave off the end.
-	 */
-	if (wloffset + wlsize < wlavailable) {
-		if (wlavailable > 0) {
-			rarrow = 1;
-			wlavailable--;
-		}
-		wlwidth = wlavailable;
-	} else {
-		/*
-		 * Work out how many characters we need to omit from the
-		 * start. There are wlavailable characters to fill, and
-		 * wloffset + wlsize must be the last. So, the start character
-		 * is wloffset + wlsize - wlavailable.
-		 */
-		if (wlavailable > 0) {
-			larrow = 1;
-			wlavailable--;
-		}
-
-		wlstart = wloffset + wlsize - wlavailable;
-		if (wlavailable > 0 && wlwidth > wlstart + wlavailable + 1) {
-			rarrow = 1;
-			wlstart++;
-			wlavailable--;
-		}
-		wlwidth = wlavailable;
-	}
-
-	/* Bail if anything is now too small too. */
-	if (wlwidth == 0 || wlavailable == 0) {
-		screen_free(&window_list);
-		goto out;
-	}
-
-	/*
-	 * Now the start position is known, work out the state of the left and
-	 * right arrows.
-	 */
-	offset = 0;
-	RB_FOREACH(wl, winlinks, &s->windows) {
-		if (wl->flags & WINLINK_ALERTFLAGS &&
-		    larrow == 1 && offset < wlstart)
-			larrow = -1;
-
-		offset += wl->status_width;
-
-		if (wl->flags & WINLINK_ALERTFLAGS &&
-		    rarrow == 1 && offset > wlstart + wlwidth)
-			rarrow = -1;
-	}
-
-draw:
-	/* Begin drawing. */
-	screen_write_start(&ctx, NULL, &c->status.status);
-
-	/* Draw the left string and arrow. */
-	screen_write_cursormove(&ctx, 0, 0);
-	if (llen != 0)
-		screen_write_cnputs(&ctx, llen, &lgc, "%s", left);
-	if (larrow != 0) {
-		memcpy(&gc, &stdgc, sizeof gc);
-		if (larrow == -1)
-			gc.attr ^= GRID_ATTR_REVERSE;
-		screen_write_putc(&ctx, &gc, '<');
-	}
-
-	/* Draw the right string and arrow. */
-	if (rarrow != 0) {
-		screen_write_cursormove(&ctx, c->tty.sx - rlen - 1, 0);
-		memcpy(&gc, &stdgc, sizeof gc);
-		if (rarrow == -1)
-			gc.attr ^= GRID_ATTR_REVERSE;
-		screen_write_putc(&ctx, &gc, '>');
-	} else
-		screen_write_cursormove(&ctx, c->tty.sx - rlen, 0);
-	if (rlen != 0)
-		screen_write_cnputs(&ctx, rlen, &rgc, "%s", right);
-
-	/* Figure out the offset for the window list. */
-	if (llen != 0)
-		wloffset = llen;
-	else
-		wloffset = 0;
-	if (wlwidth < wlavailable) {
-		switch (options_get_number(s->options, "status-justify")) {
-		case 1:	/* centred */
-			wloffset += (wlavailable - wlwidth) / 2;
-			break;
-		case 2:	/* right */
-			wloffset += (wlavailable - wlwidth);
-			break;
-		}
-	}
-	if (larrow != 0)
-		wloffset++;
-
-	/* Copy the window list. */
-	c->wlmouse = -wloffset + wlstart;
-	screen_write_cursormove(&ctx, wloffset, 0);
-	screen_write_fast_copy(&ctx, &window_list, wlstart, 0, wlwidth, 1);
-	screen_free(&window_list);
-
-	screen_write_stop(&ctx);
-
-out:
-	free(left);
-	free(right);
-
-	if (grid_compare(c->status.status.grid, old_status.grid) == 0) {
-		screen_free(&old_status);
-		return (0);
-	}
-	screen_free(&old_status);
-	return (1);
-}
-
-/* Replace special sequences in fmt. */
-static char *
-status_replace(struct client *c, struct winlink *wl, const char *fmt, time_t t)
-{
-	struct format_tree	*ft;
-	char			*expanded;
-	u_int			 tag;
-
-	if (fmt == NULL)
-		return (xstrdup(""));
-
-	if (wl != NULL)
-		tag = FORMAT_WINDOW|wl->window->id;
-	else
-		tag = FORMAT_NONE;
+	/* Create format tree. */
+	flags = FORMAT_STATUS;
 	if (c->flags & CLIENT_STATUSFORCE)
-		ft = format_create(c, NULL, tag, FORMAT_STATUS|FORMAT_FORCE);
-	else
-		ft = format_create(c, NULL, tag, FORMAT_STATUS);
-	format_defaults(ft, c, NULL, wl, NULL);
+		flags |= FORMAT_FORCE;
+	ft = format_create(c, NULL, FORMAT_NONE, flags);
+	format_defaults(ft, c, NULL, NULL, NULL);
 
-	expanded = format_expand_time(ft, fmt, t);
+	/* Write the status lines. */
+	o = options_get(s->options, "status-format");
+	if (o == NULL)
+		screen_write_clearscreen(&ctx, gc.bg);
+	else {
+		for (i = 0; i < lines; i++) {
+			screen_write_cursormove(&ctx, 0, i, 0);
 
-	format_free(ft);
-	return (expanded);
-}
+			fmt = options_array_get(o, i);
+			if (fmt == NULL) {
+				screen_write_clearline(&ctx, gc.bg);
+				continue;
+			}
+			sle = &sl->entries[i];
 
-/* Return winlink status line entry and adjust gc as necessary. */
-static char *
-status_print(struct client *c, struct winlink *wl, time_t t,
-    struct grid_cell *gc)
-{
-	struct options	*oo = wl->window->options;
-	struct session	*s = c->session;
-	const char	*fmt;
-	char   		*text;
+			expanded = format_expand_time(ft, fmt);
+			if (!force &&
+			    sle->expanded != NULL &&
+			    strcmp(expanded, sle->expanded) == 0) {
+				free(expanded);
+				continue;
+			}
+			changed = 1;
 
-	style_apply_update(gc, oo, "window-status-style");
-	fmt = options_get_string(oo, "window-status-format");
-	if (wl == s->curw) {
-		style_apply_update(gc, oo, "window-status-current-style");
-		fmt = options_get_string(oo, "window-status-current-format");
+			screen_write_clearline(&ctx, gc.bg);
+			status_free_ranges(&sle->ranges);
+			format_draw(&ctx, &gc, width, expanded, &sle->ranges);
+
+			free(sle->expanded);
+			sle->expanded = expanded;
+		}
 	}
-	if (wl == TAILQ_FIRST(&s->lastw))
-		style_apply_update(gc, oo, "window-status-last-style");
+	screen_write_stop(&ctx);
 
-	if (wl->flags & WINLINK_BELL)
-		style_apply_update(gc, oo, "window-status-bell-style");
-	else if (wl->flags & (WINLINK_ACTIVITY|WINLINK_SILENCE))
-		style_apply_update(gc, oo, "window-status-activity-style");
+	/* Free the format tree. */
+	format_free(ft);
 
-	text = status_replace(c, wl, fmt, t);
-	return (text);
+	/* Return if the status line has changed. */
+	log_debug("%s exit: force=%d, changed=%d", __func__, force, changed);
+	return (force || changed);
 }
 
 /* Set a status line message. */
@@ -589,13 +413,7 @@ status_message_set(struct client *c, const char *fmt, ...)
 	int		delay;
 
 	status_message_clear(c);
-
-	if (c->status.old_status == NULL) {
-		c->status.old_status = xmalloc(sizeof *c->status.old_status);
-		memcpy(c->status.old_status, &c->status.status,
-		    sizeof *c->status.old_status);
-		screen_init(&c->status.status, c->tty.sx, 1, 0);
-	}
+	status_push_screen(c);
 
 	va_start(ap, fmt);
 	xvasprintf(&c->message_string, fmt, ap);
@@ -615,7 +433,7 @@ status_message_set(struct client *c, const char *fmt, ...)
 	}
 
 	c->tty.flags |= (TTY_NOCURSOR|TTY_FREEZE);
-	c->flags |= CLIENT_STATUS;
+	c->flags |= CLIENT_REDRAWSTATUS;
 }
 
 /* Clear status line message. */
@@ -630,9 +448,9 @@ status_message_clear(struct client *c)
 
 	if (c->prompt_string == NULL)
 		c->tty.flags &= ~(TTY_NOCURSOR|TTY_FREEZE);
-	c->flags |= CLIENT_REDRAW; /* screen was frozen and may have changed */
+	c->flags |= CLIENT_ALLREDRAWFLAGS; /* was frozen and may have changed */
 
-	screen_reinit(&c->status.status);
+	status_pop_screen(c);
 }
 
 /* Clear status line message after timer expires. */
@@ -648,23 +466,22 @@ status_message_callback(__unused int fd, __unused short event, void *data)
 int
 status_message_redraw(struct client *c)
 {
-	struct screen_write_ctx		ctx;
-	struct session		       *s = c->session;
-	struct screen		        old_status;
-	size_t			        len;
-	struct grid_cell		gc;
-	u_int				lines, offset;
+	struct status_line	*sl = &c->status;
+	struct screen_write_ctx	 ctx;
+	struct session		*s = c->session;
+	struct screen		 old_screen;
+	size_t			 len;
+	u_int			 lines, offset;
+	struct grid_cell	 gc;
 
 	if (c->tty.sx == 0 || c->tty.sy == 0)
 		return (0);
-	memcpy(&old_status, &c->status.status, sizeof old_status);
+	memcpy(&old_screen, sl->active, sizeof old_screen);
 
-	lines = status_line_size(c->session);
-	if (lines <= 1) {
+	lines = status_line_size(c);
+	if (lines <= 1)
 		lines = 1;
-		screen_init(&c->status.status, c->tty.sx, 1, 0);
-	} else
-		screen_init(&c->status.status, c->tty.sx, lines, 0);
+	screen_init(sl->active, c->tty.sx, lines, 0);
 
 	len = screen_write_strlen("%s", c->message_string);
 	if (len > c->tty.sx)
@@ -672,19 +489,20 @@ status_message_redraw(struct client *c)
 
 	style_apply(&gc, s->options, "message-style");
 
-	screen_write_start(&ctx, NULL, &c->status.status);
-	screen_write_cursormove(&ctx, 0, 0);
-	for (offset = 0; offset < lines * c->tty.sx; offset++)
+	screen_write_start(&ctx, NULL, sl->active);
+	screen_write_fast_copy(&ctx, &sl->screen, 0, 0, c->tty.sx, lines - 1);
+	screen_write_cursormove(&ctx, 0, lines - 1, 0);
+	for (offset = 0; offset < c->tty.sx; offset++)
 		screen_write_putc(&ctx, &gc, ' ');
-	screen_write_cursormove(&ctx, 0, lines - 1);
+	screen_write_cursormove(&ctx, 0, lines - 1, 0);
 	screen_write_nputs(&ctx, len, &gc, "%s", c->message_string);
 	screen_write_stop(&ctx);
 
-	if (grid_compare(c->status.status.grid, old_status.grid) == 0) {
-		screen_free(&old_status);
+	if (grid_compare(sl->active->grid, old_screen.grid) == 0) {
+		screen_free(&old_screen);
 		return (0);
 	}
-	screen_free(&old_status);
+	screen_free(&old_screen);
 	return (1);
 }
 
@@ -694,31 +512,23 @@ status_prompt_set(struct client *c, const char *msg, const char *input,
     prompt_input_cb inputcb, prompt_free_cb freecb, void *data, int flags)
 {
 	struct format_tree	*ft;
-	time_t			 t;
 	char			*tmp, *cp;
 
 	ft = format_create(c, NULL, FORMAT_NONE, 0);
 	format_defaults(ft, c, NULL, NULL, NULL);
-	t = time(NULL);
 
 	if (input == NULL)
 		input = "";
 	if (flags & PROMPT_NOFORMAT)
 		tmp = xstrdup(input);
 	else
-		tmp = format_expand_time(ft, input, t);
+		tmp = format_expand_time(ft, input);
 
 	status_message_clear(c);
 	status_prompt_clear(c);
+	status_push_screen(c);
 
-	if (c->status.old_status == NULL) {
-		c->status.old_status = xmalloc(sizeof *c->status.old_status);
-		memcpy(c->status.old_status, &c->status.status,
-		    sizeof *c->status.old_status);
-		screen_init(&c->status.status, c->tty.sx, 1, 0);
-	}
-
-	c->prompt_string = format_expand_time(ft, msg, t);
+	c->prompt_string = format_expand_time(ft, msg);
 
 	c->prompt_buffer = utf8_fromcstr(tmp);
 	c->prompt_index = utf8_strlen(c->prompt_buffer);
@@ -734,7 +544,7 @@ status_prompt_set(struct client *c, const char *msg, const char *input,
 
 	if (~flags & PROMPT_INCREMENTAL)
 		c->tty.flags |= (TTY_NOCURSOR|TTY_FREEZE);
-	c->flags |= CLIENT_STATUS;
+	c->flags |= CLIENT_REDRAWSTATUS;
 
 	if ((flags & PROMPT_INCREMENTAL) && *tmp != '\0') {
 		xasprintf(&cp, "=%s", tmp);
@@ -762,10 +572,13 @@ status_prompt_clear(struct client *c)
 	free(c->prompt_buffer);
 	c->prompt_buffer = NULL;
 
-	c->tty.flags &= ~(TTY_NOCURSOR|TTY_FREEZE);
-	c->flags |= CLIENT_REDRAW; /* screen was frozen and may have changed */
+	free(c->prompt_saved);
+	c->prompt_saved = NULL;
 
-	screen_reinit(&c->status.status);
+	c->tty.flags &= ~(TTY_NOCURSOR|TTY_FREEZE);
+	c->flags |= CLIENT_ALLREDRAWFLAGS; /* was frozen and may have changed */
+
+	status_pop_screen(c);
 }
 
 /* Update status line prompt with a new prompt string. */
@@ -773,17 +586,15 @@ void
 status_prompt_update(struct client *c, const char *msg, const char *input)
 {
 	struct format_tree	*ft;
-	time_t			 t;
 	char			*tmp;
 
 	ft = format_create(c, NULL, FORMAT_NONE, 0);
 	format_defaults(ft, c, NULL, NULL, NULL);
 
-	t = time(NULL);
-	tmp = format_expand_time(ft, input, t);
+	tmp = format_expand_time(ft, input);
 
 	free(c->prompt_string);
-	c->prompt_string = format_expand_time(ft, msg, t);
+	c->prompt_string = format_expand_time(ft, msg);
 
 	free(c->prompt_buffer);
 	c->prompt_buffer = utf8_fromcstr(tmp);
@@ -791,7 +602,7 @@ status_prompt_update(struct client *c, const char *msg, const char *input)
 
 	c->prompt_hindex = 0;
 
-	c->flags |= CLIENT_STATUS;
+	c->flags |= CLIENT_REDRAWSTATUS;
 
 	free(tmp);
 	format_free(ft);
@@ -801,23 +612,22 @@ status_prompt_update(struct client *c, const char *msg, const char *input)
 int
 status_prompt_redraw(struct client *c)
 {
+	struct status_line	*sl = &c->status;
 	struct screen_write_ctx	 ctx;
 	struct session		*s = c->session;
-	struct screen		 old_status;
-	u_int			 i, offset, left, start, pcursor, pwidth, width;
-	u_int			 lines;
+	struct screen		 old_screen;
+	u_int			 i, lines, offset, left, start, width;
+	u_int			 pcursor, pwidth;
 	struct grid_cell	 gc, cursorgc;
 
 	if (c->tty.sx == 0 || c->tty.sy == 0)
 		return (0);
-	memcpy(&old_status, &c->status.status, sizeof old_status);
+	memcpy(&old_screen, sl->active, sizeof old_screen);
 
-	lines = status_line_size(c->session);
-	if (lines <= 1) {
+	lines = status_line_size(c);
+	if (lines <= 1)
 		lines = 1;
-		screen_init(&c->status.status, c->tty.sx, 1, 0);
-	} else
-		screen_init(&c->status.status, c->tty.sx, lines, 0);
+	screen_init(sl->active, c->tty.sx, lines, 0);
 
 	if (c->prompt_mode == PROMPT_COMMAND)
 		style_apply(&gc, s->options, "message-command-style");
@@ -831,13 +641,14 @@ status_prompt_redraw(struct client *c)
 	if (start > c->tty.sx)
 		start = c->tty.sx;
 
-	screen_write_start(&ctx, NULL, &c->status.status);
-	screen_write_cursormove(&ctx, 0, 0);
-	for (offset = 0; offset < lines * c->tty.sx; offset++)
+	screen_write_start(&ctx, NULL, sl->active);
+	screen_write_fast_copy(&ctx, &sl->screen, 0, 0, c->tty.sx, lines - 1);
+	screen_write_cursormove(&ctx, 0, lines - 1, 0);
+	for (offset = 0; offset < c->tty.sx; offset++)
 		screen_write_putc(&ctx, &gc, ' ');
-	screen_write_cursormove(&ctx, 0, 0);
+	screen_write_cursormove(&ctx, 0, lines - 1, 0);
 	screen_write_nputs(&ctx, start, &gc, "%s", c->prompt_string);
-	screen_write_cursormove(&ctx, start, 0);
+	screen_write_cursormove(&ctx, start, lines - 1, 0);
 
 	left = c->tty.sx - start;
 	if (left == 0)
@@ -877,18 +688,17 @@ status_prompt_redraw(struct client *c)
 			screen_write_cell(&ctx, &cursorgc);
 		}
 	}
-	if (c->status.status.cx < screen_size_x(&c->status.status) &&
-	    c->prompt_index >= i)
+	if (sl->active->cx < screen_size_x(sl->active) && c->prompt_index >= i)
 		screen_write_putc(&ctx, &cursorgc, ' ');
 
 finished:
 	screen_write_stop(&ctx);
 
-	if (grid_compare(c->status.status.grid, old_status.grid) == 0) {
-		screen_free(&old_status);
+	if (grid_compare(sl->active->grid, old_screen.grid) == 0) {
+		screen_free(&old_screen);
 		return (0);
 	}
-	screen_free(&old_status);
+	screen_free(&old_screen);
 	return (1);
 }
 
@@ -938,7 +748,7 @@ status_prompt_translate_key(struct client *c, key_code key, key_code *new_key)
 			return (1);
 		case '\033': /* Escape */
 			c->prompt_mode = PROMPT_COMMAND;
-			c->flags |= CLIENT_STATUS;
+			c->flags |= CLIENT_REDRAWSTATUS;
 			return (0);
 		}
 		*new_key = key;
@@ -952,17 +762,17 @@ status_prompt_translate_key(struct client *c, key_code key, key_code *new_key)
 	case 's':
 	case 'a':
 		c->prompt_mode = PROMPT_ENTRY;
-		c->flags |= CLIENT_STATUS;
+		c->flags |= CLIENT_REDRAWSTATUS;
 		break; /* switch mode and... */
 	case 'S':
 		c->prompt_mode = PROMPT_ENTRY;
-		c->flags |= CLIENT_STATUS;
+		c->flags |= CLIENT_REDRAWSTATUS;
 		*new_key = '\025'; /* C-u */
 		return (1);
 	case 'i':
 	case '\033': /* Escape */
 		c->prompt_mode = PROMPT_ENTRY;
-		c->flags |= CLIENT_STATUS;
+		c->flags |= CLIENT_REDRAWSTATUS;
 		return (0);
 	}
 
@@ -1055,6 +865,7 @@ status_prompt_key(struct client *c, key_code key)
 		free(s);
 		return (1);
 	}
+	key &= ~KEYC_XTERM;
 
 	keys = options_get_number(c->session->options, "status-keys");
 	if (keys == MODEKEY_VI) {
@@ -1212,6 +1023,12 @@ process_key:
 			}
 		}
 
+		free(c->prompt_saved);
+		c->prompt_saved = xcalloc(sizeof *c->prompt_buffer,
+		    (c->prompt_index - idx) + 1);
+		memcpy(c->prompt_saved, c->prompt_buffer + idx,
+		    (c->prompt_index - idx) * sizeof *c->prompt_buffer);
+
 		memmove(c->prompt_buffer + idx,
 		    c->prompt_buffer + c->prompt_index,
 		    (size + 1 - c->prompt_index) *
@@ -1222,6 +1039,7 @@ process_key:
 
 		goto changed;
 	case 'f'|KEYC_ESCAPE:
+	case KEYC_RIGHT|KEYC_CTRL:
 		ws = options_get_string(oo, "word-separators");
 
 		/* Find a word. */
@@ -1245,6 +1063,7 @@ process_key:
 
 		goto changed;
 	case 'b'|KEYC_ESCAPE:
+	case KEYC_LEFT|KEYC_CTRL:
 		ws = options_get_string(oo, "word-separators");
 
 		/* Find a non-separator. */
@@ -1283,22 +1102,28 @@ process_key:
 		c->prompt_index = utf8_strlen(c->prompt_buffer);
 		goto changed;
 	case '\031': /* C-y */
-		if ((pb = paste_get_top(NULL)) == NULL)
-			break;
-		bufdata = paste_buffer_data(pb, &bufsize);
-		for (n = 0; n < bufsize; n++) {
-			ch = (u_char)bufdata[n];
-			if (ch < 32 || ch >= 127)
+		if (c->prompt_saved != NULL) {
+			ud = c->prompt_saved;
+			n = utf8_strlen(c->prompt_saved);
+		} else {
+			if ((pb = paste_get_top(NULL)) == NULL)
 				break;
+			bufdata = paste_buffer_data(pb, &bufsize);
+			for (n = 0; n < bufsize; n++) {
+				ch = (u_char)bufdata[n];
+				if (ch < 32 || ch >= 127)
+					break;
+			}
+			ud = xreallocarray(NULL, n, sizeof *ud);
+			for (idx = 0; idx < n; idx++)
+				utf8_set(&ud[idx], bufdata[idx]);
 		}
 
 		c->prompt_buffer = xreallocarray(c->prompt_buffer, size + n + 1,
 		    sizeof *c->prompt_buffer);
 		if (c->prompt_index == size) {
-			for (idx = 0; idx < n; idx++) {
-				ud = &c->prompt_buffer[c->prompt_index + idx];
-				utf8_set(ud, bufdata[idx]);
-			}
+			memcpy(c->prompt_buffer + c->prompt_index, ud,
+			    n * sizeof *c->prompt_buffer);
 			c->prompt_index += n;
 			c->prompt_buffer[c->prompt_index].size = 0;
 		} else {
@@ -1306,12 +1131,13 @@ process_key:
 			    c->prompt_buffer + c->prompt_index,
 			    (size + 1 - c->prompt_index) *
 			    sizeof *c->prompt_buffer);
-			for (idx = 0; idx < n; idx++) {
-				ud = &c->prompt_buffer[c->prompt_index + idx];
-				utf8_set(ud, bufdata[idx]);
-			}
+			memcpy(c->prompt_buffer + c->prompt_index, ud,
+			    n * sizeof *c->prompt_buffer);
 			c->prompt_index += n;
 		}
+
+		if (ud != c->prompt_saved)
+			free(ud);
 		goto changed;
 	case '\024': /* C-t */
 		idx = c->prompt_index;
@@ -1357,7 +1183,7 @@ process_key:
 		goto append_key;
 	}
 
-	c->flags |= CLIENT_STATUS;
+	c->flags |= CLIENT_REDRAWSTATUS;
 	return (0);
 
 append_key:
@@ -1392,7 +1218,7 @@ append_key:
 	}
 
 changed:
-	c->flags |= CLIENT_STATUS;
+	c->flags |= CLIENT_REDRAWSTATUS;
 	if (c->prompt_flags & PROMPT_INCREMENTAL) {
 		s = utf8_tocstr(c->prompt_buffer);
 		xasprintf(&cp, "%c%s", prefix, s);
@@ -1456,12 +1282,17 @@ status_prompt_add_history(const char *line)
 }
 
 /* Build completion list. */
-static const char **
+char **
 status_prompt_complete_list(u_int *size, const char *s)
 {
-	const char				**list = NULL, **layout;
+	char					**list = NULL;
+	const char				**layout, *value, *cp;
 	const struct cmd_entry			**cmdent;
 	const struct options_table_entry	 *oe;
+	u_int					  idx;
+	size_t					  slen = strlen(s), valuelen;
+	struct options_entry			 *o;
+	struct options_array_item		 *a;
 	const char				 *layouts[] = {
 		"even-horizontal", "even-vertical", "main-horizontal",
 		"main-vertical", "tiled", NULL
@@ -1469,29 +1300,50 @@ status_prompt_complete_list(u_int *size, const char *s)
 
 	*size = 0;
 	for (cmdent = cmd_table; *cmdent != NULL; cmdent++) {
-		if (strncmp((*cmdent)->name, s, strlen(s)) == 0) {
+		if (strncmp((*cmdent)->name, s, slen) == 0) {
 			list = xreallocarray(list, (*size) + 1, sizeof *list);
-			list[(*size)++] = (*cmdent)->name;
+			list[(*size)++] = xstrdup((*cmdent)->name);
 		}
 	}
 	for (oe = options_table; oe->name != NULL; oe++) {
-		if (strncmp(oe->name, s, strlen(s)) == 0) {
+		if (strncmp(oe->name, s, slen) == 0) {
 			list = xreallocarray(list, (*size) + 1, sizeof *list);
-			list[(*size)++] = oe->name;
+			list[(*size)++] = xstrdup(oe->name);
 		}
 	}
 	for (layout = layouts; *layout != NULL; layout++) {
-		if (strncmp(*layout, s, strlen(s)) == 0) {
+		if (strncmp(*layout, s, slen) == 0) {
 			list = xreallocarray(list, (*size) + 1, sizeof *list);
-			list[(*size)++] = *layout;
+			list[(*size)++] = xstrdup(*layout);
 		}
 	}
+	o = options_get_only(global_options, "command-alias");
+	if (o != NULL) {
+		a = options_array_first(o);
+		while (a != NULL) {
+			value = options_array_item_value(a);;
+			if (value == NULL || (cp = strchr(value, '=')) == NULL)
+				goto next;
+
+			valuelen = cp - value;
+			if (slen > valuelen || strncmp(value, s, slen) != 0)
+				goto next;
+
+			list = xreallocarray(list, (*size) + 1, sizeof *list);
+			list[(*size)++] = xstrndup(value, valuelen);
+
+		next:
+			a = options_array_next(a);
+		}
+	}
+	for (idx = 0; idx < (*size); idx++)
+		log_debug("complete %u: %s", idx, list[idx]);
 	return (list);
 }
 
 /* Find longest prefix. */
 static char *
-status_prompt_complete_prefix(const char **list, u_int size)
+status_prompt_complete_prefix(char **list, u_int size)
 {
 	char	 *out;
 	u_int	  i;
@@ -1514,7 +1366,8 @@ status_prompt_complete_prefix(const char **list, u_int size)
 static char *
 status_prompt_complete(struct session *session, const char *s)
 {
-	const char	**list = NULL, *colon;
+	char		**list = NULL;
+	const char	 *colon;
 	u_int		  size = 0, i;
 	struct session	 *s_loop;
 	struct winlink	 *wl;
@@ -1533,6 +1386,8 @@ status_prompt_complete(struct session *session, const char *s)
 			xasprintf(&out, "%s ", list[0]);
 		else
 			out = status_prompt_complete_prefix(list, size);
+		for (i = 0; i < size; i++)
+			free(list[i]);
 		free(list);
 		return (out);
 	}
