@@ -60,6 +60,17 @@ static u_int	next_window_pane_id;
 static u_int	next_window_id;
 static u_int	next_active_point;
 
+/* List of window modes. */
+const struct window_mode *all_window_modes[] = {
+	&window_buffer_mode,
+	&window_client_mode,
+	&window_clock_mode,
+	&window_copy_mode,
+	&window_tree_mode,
+	&window_view_mode,
+	NULL
+};
+
 static void	window_destroy(struct window *);
 
 static struct window_pane *window_pane_create(struct window *, u_int, u_int,
@@ -204,7 +215,6 @@ winlink_remove(struct winlinks *wwl, struct winlink *wl)
 	}
 
 	RB_REMOVE(winlinks, wwl, wl);
-	free(wl->status_text);
 	free(wl);
 }
 
@@ -378,6 +388,8 @@ window_destroy(struct window *w)
 
 	if (event_initialized(&w->alerts_timer))
 		evtimer_del(&w->alerts_timer);
+	if (event_initialized(&w->offset_timer))
+		event_del(&w->offset_timer);
 
 	options_free(w->options);
 
@@ -456,17 +468,9 @@ window_set_active_pane(struct window *w, struct window_pane *wp)
 		return (0);
 	w->last = w->active;
 	w->active = wp;
-	while (!window_pane_visible(w->active)) {
-		w->active = TAILQ_PREV(w->active, window_panes, entry);
-		if (w->active == NULL)
-			w->active = TAILQ_LAST(&w->panes, window_panes);
-		if (w->active == wp) {
-			notify_window("window-pane-changed", w);
-			return (1);
-		}
-	}
 	w->active->active_point = next_active_point++;
 	w->active->flags |= PANE_CHANGED;
+	tty_update_window_offset(w);
 	notify_window("window-pane-changed", w);
 	return (1);
 }
@@ -474,7 +478,7 @@ window_set_active_pane(struct window *w, struct window_pane *wp)
 void
 window_redraw_active_switch(struct window *w, struct window_pane *wp)
 {
-	const struct grid_cell	*gc;
+	struct style	*sy;
 
 	if (wp == w->active)
 		return;
@@ -483,21 +487,21 @@ window_redraw_active_switch(struct window *w, struct window_pane *wp)
 	 * If window-style and window-active-style are the same, we don't need
 	 * to redraw panes when switching active panes.
 	 */
-	gc = options_get_style(w->options, "window-active-style");
-	if (style_equal(gc, options_get_style(w->options, "window-style")))
+	sy = options_get_style(w->options, "window-active-style");
+	if (style_equal(sy, options_get_style(w->options, "window-style")))
 		return;
 
 	/*
 	 * If the now active or inactive pane do not have a custom style or if
 	 * the palette is different, they need to be redrawn.
 	 */
-	if (window_pane_get_palette(w->active, w->active->colgc.fg) != -1 ||
-	    window_pane_get_palette(w->active, w->active->colgc.bg) != -1 ||
-	    style_equal(&grid_default_cell, &w->active->colgc))
+	if (window_pane_get_palette(w->active, w->active->style.gc.fg) != -1 ||
+	    window_pane_get_palette(w->active, w->active->style.gc.bg) != -1 ||
+	    style_is_default(&w->active->style))
 		w->active->flags |= PANE_REDRAW;
-	if (window_pane_get_palette(wp, wp->colgc.fg) != -1 ||
-	    window_pane_get_palette(wp, wp->colgc.bg) != -1 ||
-	    style_equal(&grid_default_cell, &wp->colgc))
+	if (window_pane_get_palette(wp, wp->style.gc.fg) != -1 ||
+	    window_pane_get_palette(wp, wp->style.gc.bg) != -1 ||
+	    style_is_default(&wp->style))
 		wp->flags |= PANE_REDRAW;
 }
 
@@ -561,9 +565,6 @@ window_zoom(struct window_pane *wp)
 	if (w->flags & WINDOW_ZOOMED)
 		return (-1);
 
-	if (!window_pane_visible(wp))
-		return (-1);
-
 	if (window_count_panes(w) == 1)
 		return (-1);
 
@@ -600,7 +601,7 @@ window_unzoom(struct window *w)
 		wp->layout_cell = wp->saved_layout_cell;
 		wp->saved_layout_cell = NULL;
 	}
-	layout_fix_panes(w, w->sx, w->sy);
+	layout_fix_panes(w);
 	notify_window("window-layout-changed", w);
 
 	return (0);
@@ -817,8 +818,7 @@ window_pane_create(struct window *w, u_int sx, u_int sy, u_int hlimit)
 	wp->fd = -1;
 	wp->event = NULL;
 
-	wp->mode = NULL;
-	wp->modeprefix = 1;
+	TAILQ_INIT(&wp->modes);
 
 	wp->layout_cell = NULL;
 
@@ -834,7 +834,7 @@ window_pane_create(struct window *w, u_int sx, u_int sy, u_int hlimit)
 
 	wp->saved_grid = NULL;
 
-	memcpy(&wp->colgc, &grid_default_cell, sizeof wp->colgc);
+	style_set(&wp->style, &grid_default_cell);
 
 	screen_init(&wp->base, sx, sy, hlimit);
 	wp->screen = &wp->base;
@@ -852,7 +852,7 @@ window_pane_create(struct window *w, u_int sx, u_int sy, u_int hlimit)
 static void
 window_pane_destroy(struct window_pane *wp)
 {
-	window_pane_reset_mode(wp);
+	window_pane_reset_mode_all(wp);
 	free(wp->searchstr);
 
 	if (wp->fd != -1) {
@@ -865,6 +865,8 @@ window_pane_destroy(struct window_pane *wp)
 
 	input_free(wp);
 
+	screen_free(&wp->status_screen);
+
 	screen_free(&wp->base);
 	if (wp->saved_grid != NULL)
 		grid_destroy(wp->saved_grid);
@@ -898,7 +900,6 @@ window_pane_spawn(struct window_pane *wp, int argc, char **argv,
 #ifdef HAVE_UTEMPTER
 	char		 s[32];
 #endif
-	int		 i;
 	sigset_t	 set, oldset;
 
 	if (wp->fd != -1) {
@@ -935,6 +936,7 @@ window_pane_spawn(struct window_pane *wp, int argc, char **argv,
 	sigprocmask(SIG_BLOCK, &set, &oldset);
 	switch (wp->pid = fdforkpty(ptm_fd, &wp->fd, wp->tty, NULL, &ws)) {
 	case -1:
+		wp->event = NULL;
 		wp->fd = -1;
 
 		xasprintf(cause, "%s: %s", cmd, strerror(errno));
@@ -1020,6 +1022,11 @@ window_pane_spawn(struct window_pane *wp, int argc, char **argv,
 
 	wp->event = bufferevent_new(wp->fd, window_pane_read_callback, NULL,
 	    window_pane_error_callback, wp);
+	if (wp->event == NULL)
+		fatalx("out of memory");
+
+	wp->pipe_off = 0;
+	wp->flags &= ~PANE_EXITED;
 
 	bufferevent_setwatermark(wp->event, EV_READ, 0, READ_SIZE);
 	bufferevent_enable(wp->event, EV_READ|EV_WRITE);
@@ -1065,14 +1072,18 @@ window_pane_error_callback(__unused struct bufferevent *bufev,
 void
 window_pane_resize(struct window_pane *wp, u_int sx, u_int sy)
 {
+	struct window_mode_entry	*wme;
+
 	if (sx == wp->sx && sy == wp->sy)
 		return;
 	wp->sx = sx;
 	wp->sy = sy;
 
 	screen_resize(&wp->base, sx, sy, wp->saved_grid == NULL);
-	if (wp->mode != NULL)
-		wp->mode->resize(wp, sx, sy);
+
+	wme = TAILQ_FIRST(&wp->modes);
+	if (wme != NULL && wme->mode->resize != NULL)
+		wme->mode->resize(wme, sx, sy);
 
 	wp->flags |= PANE_RESIZE;
 }
@@ -1193,7 +1204,7 @@ window_pane_reset_palette(struct window_pane *wp)
 }
 
 int
-window_pane_get_palette(const struct window_pane *wp, int c)
+window_pane_get_palette(struct window_pane *wp, int c)
 {
 	int	new;
 
@@ -1226,7 +1237,7 @@ window_pane_mode_timer(__unused int fd, __unused short events, void *arg)
 
 	if (wp->modelast < time(NULL) - WINDOW_MODE_TIMEOUT) {
 		if (ioctl(wp->fd, FIONREAD, &n) == -1 || n > 0)
-			window_pane_reset_mode(wp);
+			window_pane_reset_mode_all(wp);
 	}
 }
 
@@ -1234,39 +1245,67 @@ int
 window_pane_set_mode(struct window_pane *wp, const struct window_mode *mode,
     struct cmd_find_state *fs, struct args *args)
 {
-	struct screen	*s;
-	struct timeval	 tv = { .tv_sec = 10 };
+	struct timeval			 tv = { .tv_sec = 10 };
+	struct window_mode_entry	*wme;
 
-	if (wp->mode != NULL)
+	if (!TAILQ_EMPTY(&wp->modes) && TAILQ_FIRST(&wp->modes)->mode == mode)
 		return (1);
-	wp->mode = mode;
 
 	wp->modelast = time(NULL);
-	evtimer_set(&wp->modetimer, window_pane_mode_timer, wp);
-	evtimer_add(&wp->modetimer, &tv);
+	if (TAILQ_EMPTY(&wp->modes)) {
+		evtimer_set(&wp->modetimer, window_pane_mode_timer, wp);
+		evtimer_add(&wp->modetimer, &tv);
+	}
 
-	if ((s = wp->mode->init(wp, fs, args)) != NULL)
-		wp->screen = s;
+	TAILQ_FOREACH(wme, &wp->modes, entry) {
+		if (wme->mode == mode)
+			break;
+	}
+	if (wme != NULL) {
+		TAILQ_REMOVE(&wp->modes, wme, entry);
+		TAILQ_INSERT_HEAD(&wp->modes, wme, entry);
+	} else {
+		wme = xcalloc(1, sizeof *wme);
+		wme->wp = wp;
+		wme->mode = mode;
+		wme->prefix = 1;
+		TAILQ_INSERT_HEAD(&wp->modes, wme, entry);
+		wme->screen = wme->mode->init(wme, fs, args);
+	}
+
+	wp->screen = wme->screen;
 	wp->flags |= (PANE_REDRAW|PANE_CHANGED);
 
 	server_status_window(wp->window);
 	notify_pane("pane-mode-changed", wp);
+
 	return (0);
 }
 
 void
 window_pane_reset_mode(struct window_pane *wp)
 {
-	if (wp->mode == NULL)
+	struct window_mode_entry	*wme, *next;
+
+	if (TAILQ_EMPTY(&wp->modes))
 		return;
 
-	evtimer_del(&wp->modetimer);
+	wme = TAILQ_FIRST(&wp->modes);
+	TAILQ_REMOVE(&wp->modes, wme, entry);
+	wme->mode->free(wme);
+	free(wme);
 
-	wp->mode->free(wp);
-	wp->mode = NULL;
-	wp->modeprefix = 1;
-
-	wp->screen = &wp->base;
+	next = TAILQ_FIRST(&wp->modes);
+	if (next == NULL) {
+		log_debug("%s: no next mode", __func__);
+		evtimer_del(&wp->modetimer);
+		wp->screen = &wp->base;
+	} else {
+		log_debug("%s: next mode is %s", __func__, next->mode->name);
+		wp->screen = next->screen;
+		if (next != NULL && next->mode->resize != NULL)
+			next->mode->resize(next, wp->sx, wp->sy);
+	}
 	wp->flags |= (PANE_REDRAW|PANE_CHANGED);
 
 	server_status_window(wp->window);
@@ -1274,18 +1313,27 @@ window_pane_reset_mode(struct window_pane *wp)
 }
 
 void
-window_pane_key(struct window_pane *wp, struct client *c, struct session *s,
-    key_code key, struct mouse_event *m)
+window_pane_reset_mode_all(struct window_pane *wp)
 {
-	struct window_pane	*wp2;
+	while (!TAILQ_EMPTY(&wp->modes))
+		window_pane_reset_mode(wp);
+}
+
+void
+window_pane_key(struct window_pane *wp, struct client *c, struct session *s,
+    struct winlink *wl, key_code key, struct mouse_event *m)
+{
+	struct window_mode_entry	*wme;
+	struct window_pane		*wp2;
 
 	if (KEYC_IS_MOUSE(key) && m == NULL)
 		return;
 
-	if (wp->mode != NULL) {
+	wme = TAILQ_FIRST(&wp->modes);
+	if (wme != NULL) {
 		wp->modelast = time(NULL);
-		if (wp->mode->key != NULL)
-			wp->mode->key(wp, c, s, (key & ~KEYC_XTERM), m);
+		if (wme->mode->key != NULL)
+			wme->mode->key(wme, c, s, wl, (key & ~KEYC_XTERM), m);
 		return;
 	}
 
@@ -1298,11 +1346,11 @@ window_pane_key(struct window_pane *wp, struct client *c, struct session *s,
 		return;
 	if (options_get_number(wp->window->options, "synchronize-panes")) {
 		TAILQ_FOREACH(wp2, &wp->window->panes, entry) {
-			if (wp2 == wp || wp2->mode != NULL)
-				continue;
-			if (wp2->fd == -1 || wp2->flags & PANE_INPUTOFF)
-				continue;
-			if (window_pane_visible(wp2))
+			if (wp2 != wp &&
+			    TAILQ_EMPTY(&wp2->modes) &&
+			    wp2->fd != -1 &&
+			    (~wp2->flags & PANE_INPUTOFF) &&
+			    window_pane_visible(wp2))
 				input_key(wp2, key, NULL);
 		}
 	}
@@ -1311,16 +1359,9 @@ window_pane_key(struct window_pane *wp, struct client *c, struct session *s,
 int
 window_pane_visible(struct window_pane *wp)
 {
-	struct window	*w = wp->window;
-
-	if (wp->layout_cell == NULL)
-		return (0);
-
-	if (wp->xoff >= w->sx || wp->yoff >= w->sy)
-		return (0);
-	if (wp->xoff + wp->sx > w->sx || wp->yoff + wp->sy > w->sy)
-		return (0);
-	return (1);
+	if (~wp->window->flags & WINDOW_ZOOMED)
+		return (1);
+	return (wp == wp->window->active);
 }
 
 u_int
@@ -1377,7 +1418,7 @@ window_pane_find_up(struct window_pane *wp)
 	u_int			 edge, left, right, end, size;
 	int			 status, found;
 
-	if (wp == NULL || !window_pane_visible(wp))
+	if (wp == NULL)
 		return (NULL);
 	status = options_get_number(wp->window->options, "pane-border-status");
 
@@ -1392,7 +1433,7 @@ window_pane_find_up(struct window_pane *wp)
 	right = wp->xoff + wp->sx;
 
 	TAILQ_FOREACH(next, &wp->window->panes, entry) {
-		if (next == wp || !window_pane_visible(next))
+		if (next == wp)
 			continue;
 		if (next->yoff + next->sy + 1 != edge)
 			continue;
@@ -1424,7 +1465,7 @@ window_pane_find_down(struct window_pane *wp)
 	u_int			 edge, left, right, end, size;
 	int			 status, found;
 
-	if (wp == NULL || !window_pane_visible(wp))
+	if (wp == NULL)
 		return (NULL);
 	status = options_get_number(wp->window->options, "pane-border-status");
 
@@ -1439,7 +1480,7 @@ window_pane_find_down(struct window_pane *wp)
 	right = wp->xoff + wp->sx;
 
 	TAILQ_FOREACH(next, &wp->window->panes, entry) {
-		if (next == wp || !window_pane_visible(next))
+		if (next == wp)
 			continue;
 		if (next->yoff != edge)
 			continue;
@@ -1471,7 +1512,7 @@ window_pane_find_left(struct window_pane *wp)
 	u_int			 edge, top, bottom, end, size;
 	int			 found;
 
-	if (wp == NULL || !window_pane_visible(wp))
+	if (wp == NULL)
 		return (NULL);
 
 	list = NULL;
@@ -1485,7 +1526,7 @@ window_pane_find_left(struct window_pane *wp)
 	bottom = wp->yoff + wp->sy;
 
 	TAILQ_FOREACH(next, &wp->window->panes, entry) {
-		if (next == wp || !window_pane_visible(next))
+		if (next == wp)
 			continue;
 		if (next->xoff + next->sx + 1 != edge)
 			continue;
@@ -1517,7 +1558,7 @@ window_pane_find_right(struct window_pane *wp)
 	u_int			 edge, top, bottom, end, size;
 	int			 found;
 
-	if (wp == NULL || !window_pane_visible(wp))
+	if (wp == NULL)
 		return (NULL);
 
 	list = NULL;
@@ -1531,7 +1572,7 @@ window_pane_find_right(struct window_pane *wp)
 	bottom = wp->yoff + wp->sy;
 
 	TAILQ_FOREACH(next, &wp->window->panes, entry) {
-		if (next == wp || !window_pane_visible(next))
+		if (next == wp)
 			continue;
 		if (next->xoff != edge)
 			continue;
